# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

from __future__ import annotations

import sys
from typing import ClassVar

import torch
import torch.utils.checkpoint
from onnx import ModelProto
from packaging import version
from torch.onnx import symbolic_helper

from onnxruntime.capi._pybind_state import (
    register_input_alias_function,
    register_miscellaneous_const_input,
    register_shape_inference_function,
    register_torch_autograd_function,
)
from onnxruntime.training import ortmodule
from onnxruntime.training.utils import pytorch_scalar_type_to_pytorch_dtype, pytorch_type_to_onnx_dtype

from ._custom_op_symbolic_registry import wrap_custom_export_function
from ._fallback import ORTModuleONNXModelException, wrap_exception
from ._logger import LogColor
from ._utils import get_fully_qualified_class_name, get_runtime_pytorch_version


class _SpecialCustomFunctionHandler:
    """A class to handle high priority export of torch.autograd.Function.
    `register_high_priority_handler` can be used as function decorator to register a handler for a torch.autograd.Function.
    """

    _HIGH_PRIORITY_EXPORT_HANDLER_MAP: ClassVar[dict[str, callable]] = {}

    @staticmethod
    def add_handler(func_name: str, handler: callable) -> None:
        """Add a handler for a function name.

        Args:
            func_name (str): The function name.
            handler (callable): The handler.

        """
        _SpecialCustomFunctionHandler._HIGH_PRIORITY_EXPORT_HANDLER_MAP[func_name] = handler

    @staticmethod
    def get_handler(func_name: str) -> callable | None:
        """Get the handler for a function name.

        Args:
            func_name (str): The function name.

        Returns:
            callable | None: The handler.

        """
        return _SpecialCustomFunctionHandler._HIGH_PRIORITY_EXPORT_HANDLER_MAP.get(func_name, None)


def register_high_priority_handler(func_name):
    """Register a handler for a torch.autograd.Function using its full qualified class name."""

    def symbolic_wrapper(fn):
        _SpecialCustomFunctionHandler.add_handler(func_name, fn)
        return fn

    return symbolic_wrapper


def register_custom_function_schema_supplementary(kclass: torch.autograd.Function) -> None:
    """Register schema summplementaries, for example custom shape inference function and
     alias input function for a custom autograd.Function.

    1. The signature of the shape inference function should be:
        @staticmethod
        def infer_shape(
            node: onnx.NodeProto,
            tensor_input_shapes: List[Optional[List[Union[int, str]]]],
            tensor_input_dtypes: List[torch.onnx.TensorProtoDataType],
        ) -> Tuple[List[Optional[List[Union[int, str]]]], List[torch.onnx.TensorProtoDataType]]:
            tensor_output_shapes = []
            tensor_output_dtypes = []
            ...
            return tensor_output_shapes, tensor_output_dtypes

    The tensor_input_shapes and tensor_input_dtypes are lists of shapes and dtypes of the input tensors.
    The tensor_output_shapes and tensor_output_dtypes are lists of shapes and dtypes of the output tensors.
    Be noted: we only pass in tensor inputs, and return tensor outputs, non-tensor inputs/outputs are ignored.


    2. The signature of the alias input function should be:
        @staticmethod
        def alias_input(node_proto_str: str) -> Tuple[List[int], List[int]]:
            fw_alias_map = [1, -1, -1]
            bw_alias_map = [-1, 0]
            return fw_alias_map, bw_alias_map

    The alias input function should return a tuple of two lists:
    - The first list is the forward alias map, its length is equal to the number of all outputs of the node.
    - The second list is the backward alias map, its length is equal to the number of all inputs
        (tensor and non-tensor) of the node.

    """
    kclass_name = get_fully_qualified_class_name(kclass)
    if hasattr(kclass, "infer_shape"):
        register_shape_inference_function(kclass_name, kclass.infer_shape)

    if hasattr(kclass, "alias_input"):
        register_input_alias_function(kclass_name, kclass.alias_input)


def _get_training_mode() -> bool:
    # TODO move to public API once the exporter team exposes that
    training_mode = None
    if get_runtime_pytorch_version() >= version.parse("1.12"):
        # FIXME: using private modules
        from torch.onnx import _globals

        # before https://github.com/pytorch/pytorch/commit/c8b9b6266b505328e503b12f6a42fd88c56374f9,
        # training_mode is still a bool type
        if isinstance(_globals.GLOBALS.training_mode, bool):
            training_mode = _globals.GLOBALS.training_mode
        else:
            if _globals.GLOBALS.training_mode not in [
                torch.onnx.TrainingMode.EVAL,
                torch.onnx.TrainingMode.TRAINING,
            ]:
                raise Exception(f"Unexpected training mode {_globals.GLOBALS.training_mode}")
            training_mode = _globals.GLOBALS.training_mode == torch.onnx.TrainingMode.TRAINING
    else:
        training_mode = symbolic_helper._training_mode

    return bool(training_mode)


def _export_pt_1_10(g, n, *args, **kwargs):
    """Export torch.autograd.Function in ORT PythonOp.

    Exports PythonOp (input: "n") into a graph node in "g", and registers the PythonOp's autograd.Function in ORT backend.

    Args:
        g (jit_utils.GraphContext): The graph to export to.
        n (torch._C.Node): The PythonOp node to export, its property "pyobj" can be used to retrieve the
            torch.autograd.Function class.
            https://github.com/pytorch/pytorch/blob/68cb854d73458a14684d584c25c22b17eb79dfca/torch/csrc/jit/python/python_ir.cpp#L797
        args (list): The inputs.
        kwargs (dict): The keyword arguments.

    """
    try:
        func_class = n.pyobj().__self__
        func_full_qual_name = get_fully_qualified_class_name(func_class)

        # Check if the function is handled by high priority exporter.
        hi_pri_handler = _SpecialCustomFunctionHandler.get_handler(func_full_qual_name)
        if hi_pri_handler:
            try_export = hi_pri_handler(g, n, *args, **kwargs)
            if try_export is not None:
                return try_export

        output_tensor_types = []
        output_tensor_ranks = []
        for arg in n.outputs():
            # Type of tensor's elements.
            scalar_type = pytorch_type_to_onnx_dtype(arg.type().scalarType())
            output_tensor_types.append(scalar_type)
            output_tensor_ranks.append(arg.type().dim())
        # Fall back to common exporter if not handled by high priority exporter.
        return _default_export(
            g,
            func_full_qual_name,
            func_class,
            n.cconv(),
            n.outputsSize(),
            output_tensor_types,
            output_tensor_ranks,
            *args,
            **kwargs,
        )

    except Exception as e:
        sys.stdout.flush()
        sys.stderr.flush()
        raise wrap_exception(ORTModuleONNXModelException, e)  # noqa: B904


def _default_export(
    g, func_full_qual_name, func_class, cconv, output_size, output_tensor_types, output_tensor_ranks, *args, **kwargs
):
    input_tensor_types = []
    input_tensor_ranks = []

    input_bool_scalars = []
    input_bool_scalar_positions = []

    input_int_scalars = []
    input_int_scalar_positions = []

    input_float_scalars = []
    input_float_scalar_positions = []

    input_bool_tuples = []
    input_bool_tuple_positions = []
    input_bool_tuple_begins = []

    input_int_tuples = []
    input_int_tuple_positions = []
    input_int_tuple_begins = []

    input_float_tuples = []
    input_float_tuple_positions = []
    input_float_tuple_begins = []

    input_pointer_scalars = []
    input_pointer_scalar_positions = []

    tensor_args = []
    debug_comment = ""
    assert len(args) == len(cconv), "Number of arguments does not match calling convention"

    # Encode inputs to torch.autograd.Function.
    for i, arg, call_type in zip(range(len(args)), args, cconv, strict=False):
        if call_type == "d":
            # Got a tensor variable.
            tensor_args.append(arg)
            scalar_type = pytorch_type_to_onnx_dtype(arg.type().scalarType())
            input_tensor_types.append(scalar_type)
            input_tensor_ranks.append(arg.type().dim())
            continue

        if call_type != "c":
            raise wrap_exception(
                ORTModuleONNXModelException,
                Exception(f"Unknown calling convention found: {i}. Only 'd' and 'c' are supported"),
            )

        # Got a non-tensor variable.
        # Non-tensor can't have gradient.
        if isinstance(arg, float):
            # A float.
            input_float_scalar_positions.append(i)
            input_float_scalars.append(arg)
            continue
        # bool check MUST be before int check since bool is a subclass of int
        elif isinstance(arg, bool):
            # A bool.
            input_bool_scalar_positions.append(i)
            input_bool_scalars.append(int(arg))
            continue
        elif isinstance(arg, int):
            # A int.
            input_int_scalar_positions.append(i)
            input_int_scalars.append(arg)
            continue

        is_bool_tuple = False
        is_int_tuple = False
        is_float_tuple = False
        if isinstance(arg, tuple) and len(arg) > 0:
            # bool check MUST be before int check since bool is a subclass of int.
            is_bool_tuple = all(isinstance(ele, bool) for ele in arg)
            is_int_tuple = not is_bool_tuple and all(isinstance(ele, int) for ele in arg)
            is_float_tuple = not is_bool_tuple and not is_int_tuple and all(isinstance(ele, float) for ele in arg)

        # Only support tuple of bool, int or float, for other types, handle it as a pointer.
        if is_bool_tuple:
            # A tuple of bool.
            input_bool_tuple_positions.append(i)
            input_bool_tuple_begins.append(len(input_bool_tuples))
            input_bool_tuples.extend([int(ele) for ele in arg])
            continue
        elif is_int_tuple:
            # A tuple of ints.
            input_int_tuple_positions.append(i)
            input_int_tuple_begins.append(len(input_int_tuples))
            input_int_tuples.extend(list(arg))
            continue
        elif is_float_tuple:
            # A tuple of floats.
            input_float_tuple_positions.append(i)
            input_float_tuple_begins.append(len(input_float_tuples))
            input_float_tuples.extend(list(arg))
            continue

        from onnxruntime.training.utils.hooks._statistics_subscriber import _InspectActivation

        is_inspect_activation = func_full_qual_name == get_fully_qualified_class_name(_InspectActivation)
        if is_inspect_activation and isinstance(arg, str):
            # _InspectActivation is a special case where the first argument is a string
            # that is used to determine the activation name to be inspected.
            debug_comment += arg

        # All other inputs are accessed via "pointers".
        input_pointer_scalar_positions.append(i)
        input_pointer_scalars.append(id(arg))

        # For pointer (for example, ProcessGroup passed to PythonOp) needed for PythonOp execution,
        # we append it into a global store to hold a reference (in case it is released after module exported).
        register_miscellaneous_const_input(arg)

    attrs = {
        "func_name_s": func_full_qual_name,
        "input_convention_s": cconv,
        "outputs": output_size,
        "input_tensor_types_i": input_tensor_types,
        "input_tensor_ranks_i": input_tensor_ranks,
        "output_tensor_types_i": output_tensor_types,
        "output_tensor_ranks_i": output_tensor_ranks,
        "training_mode_i": 1 if _get_training_mode() else 0,
        "comment_s": debug_comment,
    }

    if len(input_bool_scalars) > 0:
        attrs["input_bool_scalars_i"] = input_bool_scalars
        attrs["input_bool_scalar_positions_i"] = input_bool_scalar_positions
    if len(input_int_scalars) > 0:
        attrs["input_int_scalars_i"] = input_int_scalars
        attrs["input_int_scalar_positions_i"] = input_int_scalar_positions
    if len(input_float_scalars) > 0:
        attrs["input_float_scalars_f"] = input_float_scalars
        attrs["input_float_scalar_positions_i"] = input_float_scalar_positions
    if len(input_bool_tuples) > 0:
        attrs["input_bool_tuples_i"] = input_bool_tuples
        attrs["input_bool_tuple_positions_i"] = input_bool_tuple_positions
        attrs["input_bool_tuple_begins_i"] = input_bool_tuple_begins
    if len(input_int_tuples) > 0:
        attrs["input_int_tuples_i"] = input_int_tuples
        attrs["input_int_tuple_positions_i"] = input_int_tuple_positions
        attrs["input_int_tuple_begins_i"] = input_int_tuple_begins
    if len(input_float_tuples) > 0:
        attrs["input_float_tuples_f"] = input_float_tuples
        attrs["input_float_tuple_positions_i"] = input_float_tuple_positions
        attrs["input_float_tuple_begins_i"] = input_float_tuple_begins
    if len(input_pointer_scalars) > 0:
        attrs["input_pointer_scalars_i"] = input_pointer_scalars
        attrs["input_pointer_scalar_positions_i"] = input_pointer_scalar_positions

    returned_args = g.op("com.microsoft::PythonOp", *tensor_args, **attrs)

    # Register function with class names.
    register_torch_autograd_function(func_full_qual_name, func_class)

    register_custom_function_schema_supplementary(func_class)

    return returned_args


_export = wrap_custom_export_function(_export_pt_1_10)


def post_process_enabling_autograd_function(exported_model: ModelProto) -> ModelProto:
    # Loop all PythonOp, append "_ctx" as the first output.
    for index, node in enumerate(exported_model.graph.node):
        op_name_prefix = node.op_type
        if node.domain == "com.microsoft" and node.op_type == "PythonOp":
            output_names = list(node.output)
            del node.output[:]
            node.output.append(output_names[0] + "_ctx")
            node.output.extend(output_names)
            for attr in node.attribute:
                if attr.name == "func_name":
                    kclass_name = attr.s.decode("utf-8") if isinstance(attr.s, bytes) else attr.s
                    op_name_prefix = kclass_name
                    break

            node.name = f"{op_name_prefix}_id_{index}"

    return exported_model


@register_high_priority_handler("torch.utils.checkpoint.CheckpointFunction")
@register_high_priority_handler("deepspeed.checkpointing.CheckpointFunction")
def _gradient_checkpointing_export(g, n, *args, **kwargs):
    """
    Register specialized exporter for torch.autograd.Function(s) used for checkpoint activation purposes.

    Note:
        If CheckpointFunction is exported as PythonOp, the checkpoint-ed computation
        (applied on every N transformer layer) may be computed by PyTorch, not ORT.
        This situation should be especially noted for large language models such as GPT-2.

    As alternatives to using checkpoint activation:
    1. Users could leverage HierarchalORTModule to wrap the model, which only wrap exportable
    sub-nn.Module's as ORTModule. In this way, ideally, ORT could cover most of the model computation,
    other than dispatching them to PyTorch.
    2. Users could disable the check by setting the environment variable ORTMODULE_ALLOW_AUTOGRAD_CHECKPOINT=1.
    This may imply that the exported model is not fully running on ORT, users should be aware of the potential
    performance impact.
    3. Users could also leverage ORT's memory optimization feature to achieve a similar effect as checkpointing
    activations. Turn off PyTorch's checkpointing activation, then refer to env var ORTMODULE_MEMORY_OPT_LEVEL
    to enable ORT's recomputation feature.

    """
    # Check if the checkpointing activation is allowed.
    is_ckpt_activation_allowed = ortmodule._defined_from_envvar("ORTMODULE_ALLOW_AUTOGRAD_CHECKPOINT", 0) == 1
    if is_ckpt_activation_allowed is False:
        is_layerwise_recompute_enabled = ortmodule._defined_from_envvar("ORTMODULE_MEMORY_OPT_LEVEL", 0) == 1
        if not is_layerwise_recompute_enabled:
            raise Exception(
                f"{LogColor.RED}"
                "Model uses gradient checkpointing (via {func_full_qual_name}), "
                "which is not supported for export. \n"
                "Consider these alternatives:\n"
                "1) Enable ORTModule's gradient checkpointing for similar or better "
                "memory efficiency with `export ORTMODULE_MEMORY_OPT_LEVEL=1`.\n"
                "2) Allow gradient checkpointing export by setting the environment "
                "variable `ORTMODULE_ALLOW_AUTOGRAD_CHECKPOINT=1`, though subsequent "
                "execution may fail."
                "3) Replace ORTModule with HierarchalORTModule to wrap exportable "
                "sub-nn.Module's as ORTModule.\n"
                f"{LogColor.ENDC}"
            )

        # Hitting this branch means the user has enabled layerwise recompute, but _override_gradient_checkpoint didn't
        # catch the checkpointing function. This is usually because model code is importing torch.utils.checkpoint
        # earlier than ORTModule. We should tolerantly allow this case to happen.
        raise Exception(
            f"{LogColor.RED}"
            "Model uses gradient checkpointing (via {func_full_qual_name}), which is not "
            "supported for export. \n"
            "Consider these alternatives:\n"
            "1) `ORTMODULE_MEMORY_OPT_LEVEL=1` is set but checkpoint functions in the model "
            "are not overridden during onnxruntime.training.ortmodule import, consider importing "
            "onnxruntime.training.ortmodule earlier before any model code loaded.\n"
            "2) To allow gradient checkpointing export, set `ORTMODULE_ALLOW_AUTOGRAD_CHECKPOINT=1`. "
            "Subsequent execution may fail.\n"
            "3) Replace ORTModule with HierarchalORTModule to wrap exportable sub-nn.Module's as "
            "ORTModule.\n"
            f"{LogColor.ENDC}"
        )
    else:
        return None  # Let the common exporter handle the checkpointing function


@register_high_priority_handler("bitsandbytes.autograd._functions.MatMul4Bit")
def _matmul4bit_export(g, n, *args, **kwargs):
    cconv = n.cconv()
    can_converted = (
        len(cconv) >= 5
        and cconv[0] == "d"
        and cconv[1] == "d"
        and cconv[2] == "c"
        and cconv[3] == "c"
        and cconv[4] == "c"
    )
    can_converted = can_converted and (args[2] is None and args[3] is None and args[4] is not None)
    if not can_converted:
        return None

    quant_state = args[4]
    if isinstance(quant_state, list):
        # version <= 0.41.1
        absmax, shape, dtype, blocksize, compressed_stats, quant_type, data_type = quant_state
        nested = compressed_stats is not None
    else:
        # version > 0.41.1
        absmax = quant_state.absmax
        shape = quant_state.shape
        blocksize = quant_state.blocksize
        nested = quant_state.nested
        quant_type = quant_state.quant_type

    # MatMulBnb4's blocksize needs to be a power of 2 and not smaller than 16
    if blocksize < 16 or blocksize & (blocksize - 1) != 0:
        return None

    # MatMulBnb4 does not support double de-quantization (e.g. absmax is int, needs to be dequantized too)
    if nested:
        return None

    # The PyTorch linear weight shape is [out_feature, in_feature]
    in_feature = shape[1]
    out_feature = shape[0]
    if quant_type == "fp4":
        quant_type = 0
    elif quant_type == "nf4":
        quant_type = 1
    else:
        return None
    attrs = {
        "K_i": in_feature,
        "N_i": out_feature,
        "block_size_i": blocksize,
        "quant_type_i": quant_type,
        "training_mode_i": 1 if _get_training_mode() else 0,
    }

    # Make sure the quant weight can be flatten to 1D tensor safely, which com.microsoft::MatMulBnb4 requires.
    found_dim1 = any(v == 1 for v in args[1].type().sizes())
    if not found_dim1:
        return None

    absmax = g.op(
        "Constant",
        value_t=torch.tensor(absmax, dtype=pytorch_scalar_type_to_pytorch_dtype(args[0].type().scalarType())),
    )
    quant_weight = g.op(
        "Reshape", args[1], g.op("Constant", value_t=torch.tensor([-1], dtype=torch.int64))
    )  # flatten to 1D
    tensor_args = [args[0], quant_weight, absmax]
    return g.op("com.microsoft::MatMulBnb4", *tensor_args, **attrs)


class DetermisticWrapper(torch.autograd.Function):
    """
    A wrapper for run autograd function in a deterministic way. This is required for PythonOp that needs
    recompute support.
    """

    @staticmethod
    def forward(ctx, autograd_function: torch.autograd.Function, cpu_rng_state, device_rng_state, *args):
        """For normal forward run, both cpu_rng_state and device_rng_state are None.
        For recompute run, cpu_rng_state and device_rng_state are provided, and from the normal forward run results.

        If device_rng_state does not exist (in pure CPU training for example), we still return cpu_rng_state
        as device_rng_state to avoid the exporter to handle the case where we return None as forward outputs.
        """
        original_cpu_rng_state = None
        original_cuda_rng_state = None

        if cpu_rng_state is None:
            assert device_rng_state is None, "device_rng_state must be None if cpu_rng_state is None"
            cpu_rng_state = torch.get_rng_state()
            fwd_devices = list(
                {arg.device for arg in args if isinstance(arg, torch.Tensor) and arg.device.type != "cpu"}
            )
            if len(fwd_devices) > 0:
                assert len(fwd_devices) == 1, "Only support single device for now"
                assert fwd_devices[0].type == "cuda", "Only support cuda device for now"
                device_rng_state = torch.cuda.get_rng_state()
            else:
                # Pass CPU RNG state as device RNG state if device RNG state is not provided.
                # This is to workaround the tricky case where we return None|Tensor as forward outputs.
                device_rng_state = cpu_rng_state
        else:
            assert device_rng_state is not None, "device_rng_state must be provided if cpu_rng_state is provided"
            original_cpu_rng_state = torch.get_rng_state()
            torch.set_rng_state(cpu_rng_state)

            if device_rng_state.data_ptr() != cpu_rng_state.data_ptr():
                original_cuda_rng_state = torch.cuda.get_rng_state()
                torch.cuda.set_rng_state(device_rng_state)

        outputs = autograd_function.forward(ctx, *args)

        # Append the RNG states to the outputs in the beginning.
        updated_outputs = []
        updated_outputs.append(cpu_rng_state)
        updated_outputs.append(device_rng_state)
        if isinstance(outputs, torch.Tensor):
            updated_outputs.append(outputs)
        elif isinstance(outputs, tuple):
            updated_outputs.extend(outputs)
        else:
            raise ValueError(f"Unsupported outputs type: {type(outputs)}")

        ctx.autograd_function = autograd_function

        if original_cpu_rng_state is not None:
            torch.set_rng_state(original_cpu_rng_state)

        if original_cuda_rng_state is not None:
            torch.cuda.set_rng_state(original_cuda_rng_state)

        return tuple(updated_outputs)

    @staticmethod
    def backward(ctx, *grad_outputs):
        # Skip the first two RNG states grad, which should be None.
        outputs = ctx.autograd_function.backward(ctx, *grad_outputs[2:])

        updated_outputs = [None, None, None]
        if isinstance(outputs, torch.Tensor):
            updated_outputs.append(outputs)
        elif isinstance(outputs, tuple):
            updated_outputs.extend(outputs)
        else:
            raise ValueError(f"Unsupported outputs type: {type(outputs)}")

        return tuple(updated_outputs)


@register_high_priority_handler("flash_attn.bert_padding.IndexFirstAxis")
@register_high_priority_handler("flash_attn.bert_padding.IndexPutFirstAxis")
@register_high_priority_handler("flash_attn.flash_attn_interface.FlashAttnFunc")
@register_high_priority_handler("flash_attn.flash_attn_interface.FlashAttnVarlenFunc")
@register_high_priority_handler(
    "orttraining_test_ortmodule_autograd.test_determistic_pythonop_export.<locals>.TestFunction"
)
@register_high_priority_handler(
    "orttraining_test_ortmodule_api.test_layerwise_recompute_determinstic.<locals>.DropoutFunction"
)
def _determinstic_exporter(g, n, *args, **kwargs):
    """
    Export torch.autograd.Function in ORT PythonOp with deterministic wrapper. This is required for PythonOp that needs
    recompute support.

    Here, we will insert 3 inputs before the actual inputs:
    1. The first input is a constant pointer, which is not a tensor. It points to the real autograd function to execute.
    2. The second input is a tensor, which is the CPU RNG state (during export, we assign None; in memory optimizer,
      the recomputed PythonOp will take the CPU RNG state output from normal forward PythonOp node).
    3. The third input is a tensor, which is the device RNG state (during export, we assign None; in memory optimizer,
      the recomputed PythonOp will take the CUDA RNG state output from normal forward PythonOp node).

    """
    # The first input is a constant pointer, which is not a tensor. The second input rng_state is a tensor.
    cconv = "ccc" + n.cconv()
    func_class = n.pyobj().__self__
    updated_args = [func_class, None, None]
    if isinstance(args, (tuple, list)):
        updated_args.extend(args)
    else:
        updated_args.append(args)

    output_tensor_types = []
    output_tensor_ranks = []
    for arg in n.outputs():
        # Type of tensor's elements.
        scalar_type = pytorch_type_to_onnx_dtype(arg.type().scalarType())
        output_tensor_types.append(scalar_type)
        output_tensor_ranks.append(arg.type().dim())

    for _ in range(2):
        output_tensor_types.insert(0, pytorch_type_to_onnx_dtype(torch.uint8))
        output_tensor_ranks.insert(0, 1)

    func_full_qual_name = get_fully_qualified_class_name(func_class)
    default_op_outputs = _default_export(
        g,
        func_full_qual_name,
        DetermisticWrapper,
        cconv,
        n.outputsSize() + 2,
        output_tensor_types,
        output_tensor_ranks,
        *updated_args,
        **kwargs,
    )

    return default_op_outputs[2:]



import onnx
from onnx import TensorProto, helper

graph = helper.make_graph(
    [  # nodes
        # fusable, const_min_negative should be replaced
        helper.make_node("Conv", ["X", "W"], ["conv0_out"], "Conv0"),
        helper.make_node("HardSigmoid", ["conv0_out"], ["hardsigmoid0_out"], "HardSigmoid0"),
    ],
    "ConvClipFusion",  # name
    [  # inputs
        helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 1, 10, 10]),
        helper.make_tensor_value_info("W", TensorProto.FLOAT, [1, 1, 3, 3]),
    ],
    [  # outputs
        helper.make_tensor_value_info("hardsigmoid0_out", TensorProto.FLOAT, None),
    ],
)

model = helper.make_model(graph)
onnx.save(model, r"conv_hardsigmoid.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import unittest

import numpy as np
from mpi4py import MPI
from onnx import TensorProto, helper

import onnxruntime

np.random.seed(3)

comm = MPI.COMM_WORLD


def get_rank():
    return comm.Get_rank()


def get_size():
    return comm.Get_size()


def print_out(*args):
    if get_rank() == 0:
        print(*args)


local_rank = get_rank()

ORT_DTYPE = TensorProto.FLOAT16
NP_TYPE = np.float16 if ORT_DTYPE == TensorProto.FLOAT16 else np.float32
THRESHOLD_TP = 3e-2
THRESHOLD_EP = 1e-6


def create_moe_onnx_graph(
    num_rows,
    num_experts,
    local_num_experts,
    hidden_size,
    inter_size,
    fc1_experts_weights,
    fc1_experts_bias,
    fc2_experts_weights,
    fc2_experts_bias,
    fc3_experts_weights,
    local_experts_start_index=0,
    topk=2,
    normalize_routing_weights=1,
    activation_type="gelu",
    tensor_shards=1,
):
    use_sharded_moe = num_experts > local_num_experts or tensor_shards > 1
    nodes = [
        (
            helper.make_node(
                "MoE",
                [
                    "input",
                    "router_probs",
                    "fc1_experts_weights",
                    "fc1_experts_bias",
                    "fc2_experts_weights",
                    "fc2_experts_bias",
                    "fc3_experts_weights",
                ],
                ["output"],
                "MoE_0",
                k=topk,
                normalize_routing_weights=normalize_routing_weights,
                activation_type=activation_type,
                domain="com.microsoft",
            )
            if not use_sharded_moe
            else helper.make_node(
                "ShardedMoE",
                [
                    "input",
                    "router_probs",
                    "fc1_experts_weights",
                    "fc1_experts_bias",
                    "fc2_experts_weights",
                    "fc2_experts_bias",
                    "fc3_experts_weights",
                ],
                ["output"],
                "MoE_0",
                k=topk,
                normalize_routing_weights=normalize_routing_weights,
                activation_type=activation_type,
                local_experts_start_index=local_experts_start_index,
                tensor_shards=tensor_shards,
                domain="com.microsoft",
            )
        ),
    ]

    fc1_shape = [local_num_experts, hidden_size, inter_size]
    fc2_shape = [local_num_experts, inter_size, hidden_size]
    fc3_shape = fc1_shape

    initializers = [
        helper.make_tensor(
            "fc1_experts_weights",
            ORT_DTYPE,
            fc1_shape,
            fc1_experts_weights.flatten(),
            raw=False,
        ),
        helper.make_tensor(
            "fc2_experts_weights",
            ORT_DTYPE,
            fc2_shape,
            fc2_experts_weights.flatten(),
            raw=False,
        ),
        helper.make_tensor(
            "fc3_experts_weights",
            ORT_DTYPE,
            fc3_shape,
            fc3_experts_weights.flatten(),
            raw=False,
        ),
    ]

    fc1_bias_shape = [local_num_experts, inter_size]
    fc2_bias_shape = [num_experts, hidden_size]
    initializers.extend(
        [
            helper.make_tensor(
                "fc1_experts_bias",
                ORT_DTYPE,
                fc1_bias_shape,
                fc1_experts_bias.flatten().tolist(),
                raw=False,
            ),
            helper.make_tensor(
                "fc2_experts_bias",
                ORT_DTYPE,
                fc2_bias_shape,
                fc2_experts_bias.flatten().tolist(),
                raw=False,
            ),
        ]
    )

    graph_inputs = [
        helper.make_tensor_value_info("input", ORT_DTYPE, [num_rows, hidden_size]),
    ]

    graph_inputs.append(
        helper.make_tensor_value_info(
            "router_probs",
            ORT_DTYPE,
            [num_rows, num_experts],
        )
    )

    graph_outputs = [
        helper.make_tensor_value_info("output", ORT_DTYPE, [num_rows, hidden_size]),
    ]

    graph = helper.make_graph(
        nodes,
        "MoE_Graph",
        graph_inputs,
        graph_outputs,
        initializers,
    )

    model = helper.make_model(graph)
    return model.SerializeToString()


def generate_weights_and_initial_model(
    num_rows,
    num_experts,
    hidden_size,
    inter_size,
):
    s = 0.1
    fc1_experts_weights_all = np.random.normal(scale=s, size=(num_experts, hidden_size, inter_size)).astype(NP_TYPE)
    fc2_experts_weights_all = np.random.normal(scale=s, size=(num_experts, inter_size, hidden_size)).astype(NP_TYPE)
    fc3_experts_weights_all = np.random.normal(scale=s, size=(num_experts, hidden_size, inter_size)).astype(NP_TYPE)
    fc1_experts_bias_all = np.random.normal(scale=s, size=(num_experts, inter_size)).astype(NP_TYPE)
    fc2_experts_bias_all = np.random.normal(scale=s, size=(num_experts, hidden_size)).astype(NP_TYPE)

    onnx_model_full = create_moe_onnx_graph(
        num_rows,
        num_experts,
        num_experts,
        hidden_size,
        inter_size,
        fc1_experts_weights_all,
        fc1_experts_bias_all,
        fc2_experts_weights_all,
        fc2_experts_bias_all,
        fc3_experts_weights_all,
    )

    return (
        onnx_model_full,
        fc1_experts_weights_all,
        fc1_experts_bias_all,
        fc2_experts_weights_all,
        fc2_experts_bias_all,
        fc3_experts_weights_all,
    )


def run_ort_with_parity_check(
    onnx_model_full,
    onnx_model_local,
    num_rows,
    hidden_size,
    num_experts,
    inter_size,
    threshold,
):
    sess_options = onnxruntime.SessionOptions()
    cuda_provider_options = {"device_id": local_rank}
    execution_providers = [("CUDAExecutionProvider", cuda_provider_options)]

    ort_session = onnxruntime.InferenceSession(onnx_model_full, sess_options, providers=execution_providers)
    ort_session_local = onnxruntime.InferenceSession(onnx_model_local, sess_options, providers=execution_providers)

    ort_inputs = {
        ort_session.get_inputs()[0].name: np.random.rand(num_rows, hidden_size).astype(NP_TYPE),
        ort_session.get_inputs()[1].name: np.random.rand(num_rows, num_experts).astype(NP_TYPE),
    }

    output = ort_session.run(None, ort_inputs)
    sharded_output = ort_session_local.run(None, ort_inputs)

    print_out("max diff:", np.max(np.abs(output[0] - sharded_output[0])))
    assert np.allclose(output[0], sharded_output[0], atol=threshold, rtol=threshold)

    print_out(
        "hidden_size:",
        hidden_size,
        " inter_size:",
        inter_size,
        " num_experts:",
        num_experts,
        " num_rows:",
        num_rows,
        " world_size:",
        get_size(),
        " Parity: OK",
    )


def test_moe_with_tensor_parallelism(
    hidden_size,
    inter_size,
    num_experts,
    num_rows,
    threshold=THRESHOLD_TP,
):
    assert inter_size % get_size() == 0

    (
        onnx_model_full,
        fc1_experts_weights_all,
        fc1_experts_bias_all,
        fc2_experts_weights_all,
        fc2_experts_bias_all,
        fc3_experts_weights_all,
    ) = generate_weights_and_initial_model(
        num_rows,
        num_experts,
        hidden_size,
        inter_size,
    )

    def get_fc1_tensor_shards(expert_weights):
        return (
            expert_weights.reshape(-1, inter_size, hidden_size)
            .transpose(0, 2, 1)[
                :, :, local_rank * inter_size // get_size() : (local_rank + 1) * inter_size // get_size()
            ]
            .transpose(0, 2, 1)
        )

    def get_fc2_tensor_shards(expert_weights):
        return (
            expert_weights.reshape(-1, hidden_size, inter_size)
            .transpose(0, 2, 1)[
                :, local_rank * inter_size // get_size() : (local_rank + 1) * inter_size // get_size(), :
            ]
            .transpose(0, 2, 1)
        )

    fc1_experts_weights = get_fc1_tensor_shards(fc1_experts_weights_all)
    fc2_experts_weights = get_fc2_tensor_shards(fc2_experts_weights_all)
    fc3_experts_weights = get_fc1_tensor_shards(fc3_experts_weights_all)
    fc1_experts_bias = fc1_experts_bias_all[
        :, local_rank * inter_size // get_size() : (local_rank + 1) * inter_size // get_size()
    ]

    onnx_model_local = create_moe_onnx_graph(
        num_rows,
        num_experts,
        num_experts,
        hidden_size,
        inter_size // get_size(),
        fc1_experts_weights,
        fc1_experts_bias,
        fc2_experts_weights,
        fc2_experts_bias_all,
        fc3_experts_weights,
        tensor_shards=get_size(),
    )

    run_ort_with_parity_check(
        onnx_model_full,
        onnx_model_local,
        num_rows,
        hidden_size,
        num_experts,
        inter_size,
        threshold,
    )


def test_moe_with_expert_parallelism(
    hidden_size,
    inter_size,
    num_experts,
    num_rows,
    threshold=THRESHOLD_EP,
):
    local_experts_start_index = local_rank * num_experts // get_size()

    (
        onnx_model_full,
        fc1_experts_weights_all,
        fc1_experts_bias_all,
        fc2_experts_weights_all,
        fc2_experts_bias_all,
        fc3_experts_weights_all,
    ) = generate_weights_and_initial_model(
        num_rows,
        num_experts,
        hidden_size,
        inter_size,
    )

    fc1_experts_weights = fc1_experts_weights_all[
        local_experts_start_index : local_experts_start_index + num_experts // get_size(), :, :
    ]
    fc2_experts_weights = fc2_experts_weights_all[
        local_experts_start_index : local_experts_start_index + num_experts // get_size(), :, :
    ]
    fc3_experts_weights = fc3_experts_weights_all[
        local_experts_start_index : local_experts_start_index + num_experts // get_size(), :, :
    ]
    fc1_experts_bias = fc1_experts_bias_all[
        local_experts_start_index : local_experts_start_index + num_experts // get_size(), :
    ]

    onnx_model_local = create_moe_onnx_graph(
        num_rows,
        num_experts,
        num_experts // get_size(),
        hidden_size,
        inter_size,
        fc1_experts_weights,
        fc1_experts_bias,
        fc2_experts_weights,
        fc2_experts_bias_all,
        fc3_experts_weights,
        local_experts_start_index,
    )

    run_ort_with_parity_check(
        onnx_model_full,
        onnx_model_local,
        num_rows,
        hidden_size,
        num_experts,
        inter_size,
        threshold,
    )


class TestMoE(unittest.TestCase):
    def test_moe_parallelism(self):
        for hidden_size in [128, 1024]:
            for inter_size in [512, 2048]:
                for num_experts in [64]:
                    for num_rows in [1024]:
                        print_out("EP")
                        test_moe_with_expert_parallelism(
                            hidden_size,
                            inter_size,
                            num_experts,
                            num_rows,
                        )
                        print_out("TP")
                        test_moe_with_tensor_parallelism(
                            hidden_size,
                            inter_size,
                            num_experts,
                            num_rows,
                        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import torch
import torch._dynamo
from torch import nn
from torch.nn import functional as F
from torch.onnx import ExportOptions
from torch.onnx import _OrtBackend as OrtBackend
from torch.onnx import _OrtBackendOptions as OrtBackendOptions
from torch.utils import _pytree


def make_local_backend(dynamic: bool = False, use_aot_autograd: bool = False):
    ort_backend = OrtBackend(
        options=OrtBackendOptions(
            export_options=ExportOptions(
                dynamic_shapes=dynamic,
            ),
            use_aot_autograd=use_aot_autograd,
        )
    )
    return ort_backend


class TestTorchDynamoOrt(unittest.TestCase):
    """Containers of tests for TorchDynamo ORT (DORT) backend."""

    def setUp(self):
        # Make computation deterministic.
        torch.manual_seed(42)
        print(f"TestTorchDynamoOrt uses PyTorch version {torch.__version__}")

    def test_elementwise_model(self):
        torch._dynamo.reset()
        """Test DORT with a pure function."""

        def run_elementwise_model():
            # A function to test DORT.
            def elementwise_model(tensor_x: torch.Tensor):
                tensor_w = tensor_x.sigmoid()
                tensor_y = tensor_w * tensor_w + 1.5
                tensor_z = tensor_y + tensor_x
                tensor_p = tensor_z * tensor_x
                tensor_q = tensor_p.sigmoid()
                return tensor_q

            optimized_elementwise_model = torch.compile(elementwise_model, backend="onnxrt", dynamic=True)

            def run(fun, list_x):
                tensor_x = torch.tensor(list_x, dtype=torch.float32).requires_grad_()
                tensor_y = fun(tensor_x)
                tensor_y.sum().backward()
                return tensor_x, tensor_y, tensor_x.grad

            # Baseline.
            tensor_x, tensor_y, tensor_x_grad = run(elementwise_model, [-1.0, 2.0])
            # ORT result.
            tensor_x_new, tensor_y_new, tensor_x_grad_new = run(optimized_elementwise_model, [-1.0, 2.0])

            torch.testing.assert_close(tensor_x, tensor_x_new)
            torch.testing.assert_close(tensor_y, tensor_y_new)
            torch.testing.assert_close(tensor_x_grad, tensor_x_grad_new)

        # Run 5 times because ORT runs have side effects and we want to make sure
        # the code is correct with them.
        for _ in range(5):
            run_elementwise_model()

    def test_dynamo_shape_model(self):
        torch._dynamo.reset()
        """Test DORT with a pure function."""

        def run_elementwise_model():
            # A function to test DORT.
            def elementwise_model(tensor_x: torch.Tensor):
                tensor_y = tensor_x.sigmoid()
                tensor_z = tensor_y + tensor_x
                tensor_p = tensor_z * tensor_x
                tensor_q = tensor_p.sigmoid()
                return tensor_q

            # This function should only generate one graph and execute
            # it for all inputs.
            # With dynamic_shape=True, Dynamo sends FX graphs with dynamic
            # shapes (e.g., batch size is a symbol "batch" instead of a fixed
            # number) to OrtBackend.compile(...).
            optimized_elementwise_model = torch.compile(elementwise_model, backend="onnxrt", dynamic=True)

            def run(fun, seed: torch.Tensor):
                tensor_x = seed.detach().clone().requires_grad_()
                tensor_y = fun(tensor_x)
                tensor_y.sum().backward()
                return tensor_x, tensor_y, tensor_x.grad

            # Dimension changed.
            for shape in [(2, 3), (3, 4)]:
                seed = torch.rand(shape)
                # Baseline.
                tensor_x, tensor_y, tensor_x_grad = run(elementwise_model, seed)
                # ORT result.
                tensor_x_new, tensor_y_new, tensor_x_grad_new = run(optimized_elementwise_model, seed)

                torch.testing.assert_close(tensor_x, tensor_x_new)
                torch.testing.assert_close(tensor_y, tensor_y_new)
                torch.testing.assert_close(tensor_x_grad, tensor_x_grad_new)

            # Rank changed.
            for shape in [(1,), (2,), (2, 3), (2, 3, 4)]:
                seed = torch.rand(shape)
                # Baseline.
                tensor_x, tensor_y, tensor_x_grad = run(elementwise_model, seed)
                # ORT result.
                tensor_x_new, tensor_y_new, tensor_x_grad_new = run(optimized_elementwise_model, seed)

                torch.testing.assert_close(tensor_x, tensor_x_new)
                torch.testing.assert_close(tensor_y, tensor_y_new)
                torch.testing.assert_close(tensor_x_grad, tensor_x_grad_new)

        run_elementwise_model()

    def test_elementwise_model_with_dynamic_shapes_and_complicated_output_schema(self):
        torch._dynamo.reset()

        def run_elementwise_model():
            # A function to test DORT.
            def elementwise_model(tensor_x: torch.Tensor):
                tensor_y = tensor_x.sigmoid()
                tensor_z = tensor_y + tensor_x
                tensor_p = tensor_z * tensor_x
                tensor_q = tensor_p.sigmoid()
                return (tensor_q, (tensor_y, tensor_z))

            local_backend = make_local_backend(dynamic=True, use_aot_autograd=True)
            cached = local_backend._all_ort_execution_info.execution_info_per_graph_module
            # Before compilation, no graph is generated.
            assert len(cached) == 0

            # This function should only generate one graph and execute
            # it for all inputs.
            # With dynamic_shape=True, Dynamo sends FX graphs with dynamic
            # shapes (e.g., batch size is a symbol "batch" instead of a fixed
            # number) to OrtBackend.compile(...).
            @torch._dynamo.optimize(local_backend, dynamic=True)
            def optimized_elementwise_model(tensor_x: torch.Tensor):
                return elementwise_model(tensor_x)

            def run(fun, seed: torch.Tensor):
                tensor_x = seed.detach().clone().requires_grad_()
                result = fun(tensor_x)
                forward_outputs, _ = _pytree.tree_flatten(result)
                result[0].sum().backward()
                return (tensor_x, *forward_outputs, tensor_x.grad)

            # Dimension changed.
            for shape in [(2, 3), (3, 4)]:
                seed = torch.rand(shape)
                # Baseline.
                baseline_tensors = run(elementwise_model, seed)
                # ORT result.
                tensors = run(optimized_elementwise_model, seed)

                for tensor, baseline_tensor in zip(tensors, baseline_tensors, strict=False):
                    torch.testing.assert_close(tensor, baseline_tensor)

            assert len(cached.keys()) == 2, (
                "Should only see two GraphModules so far. One for forward and the other one for backward."
            )
            for value in cached.values():
                assert len(value) == 1, (
                    "One GraphModule should only be mapped to one ONNX model since "
                    "dynamic shape is enabled and input tensor's rank is unchanged."
                )

            # Rank changed.
            for shape in [(1,), (2,), (2, 3), (2, 3, 4)]:
                seed = torch.rand(shape)
                # Baseline.
                baseline_tensors = run(elementwise_model, seed)
                # ORT result.
                tensors = run(optimized_elementwise_model, seed)

                for tensor, baseline_tensor in zip(tensors, baseline_tensors, strict=False):
                    torch.testing.assert_close(tensor, baseline_tensor)

            # 4 GraphModule's respectively for
            #  - (1,)
            #  - (2,)
            #  - (2, 3), (3, 4)
            #  - (2, 3, 4)
            # Because (1,) is treated as a special dimension in Dynamo,
            # we can NOT merge (1,) and (2,). More specifically, their GraphModule's
            # are hashed to different values.
            # Another 4 GraphModule's for the corresponding backward passes.
            assert len(cached.keys()) == 8
            for value in cached.values():
                # When dynamic shape is enabled, there should be only one ONNX model
                # for inputs with the same rank.
                assert len(value) == 1

        run_elementwise_model()

    def test_elementwise_model_for_inference(self):
        torch._dynamo.reset()

        # A function to test DORT for inference (i.e., the compiled function
        # doesn't have backward pass).
        def elementwise_model(tensor_x: torch.Tensor):
            tensor_w = tensor_x.relu()
            tensor_y = tensor_w * tensor_w + 1.5
            tensor_z = tensor_y + tensor_x
            tensor_p = tensor_z * tensor_x
            tensor_q = tensor_p.relu()
            return tensor_q

        # TODO: Set use_aot_autograd=False. In order to decompose torch
        # function calls to aten ops, we need to set
        # user_aot_autograd=True because there is no decomposition in DORT
        # anymore. A long-term fix will be brining # decomposition pass back
        # into DORT.
        local_backend = make_local_backend(dynamic=True, use_aot_autograd=True)
        optimized_elementwise_model = torch.compile(elementwise_model, backend=local_backend, dynamic=True)

        def run(fun, list_x):
            tensor_x = torch.tensor(list_x, dtype=torch.float32).requires_grad_()
            tensor_y = fun(tensor_x)
            return tensor_y

        # Baseline.
        tensor_y = run(elementwise_model, [-1.0, 2.0])
        # ORT result.
        tensor_y_new = run(optimized_elementwise_model, [-1.0, 2.0])

        torch.testing.assert_close(tensor_y, tensor_y_new)

    def test_to_copy(self):
        torch._dynamo.reset()
        """Test DORT with aten::_to_copy."""

        def run_to_copy():
            # A function to test.
            def copy_copy_copy(tensor_x: torch.Tensor):
                tensor_x1 = torch.ops.aten._to_copy(tensor_x, dtype=torch.int64)
                tensor_x2 = torch.ops.aten._to_copy(tensor_x, dtype=torch.int64, device=tensor_x.device)
                tensor_x3 = torch.ops.aten._to_copy(
                    tensor_x, dtype=torch.int64, device=tensor_x.device, layout=torch.strided
                )
                return tensor_x1, tensor_x2, tensor_x3

            optimized_copy_copy_copy = torch.compile(copy_copy_copy, backend="onnxrt")

            def run(fun, list_x):
                tensor_x = torch.tensor(list_x, dtype=torch.float32)
                tensor_x1, tensor_x2, tensor_x3 = fun(tensor_x)
                return tensor_x1, tensor_x2, tensor_x3

            # Baseline.
            tensor_x, tensor_y, tensor_z = run(copy_copy_copy, [-1.0, 2.0])
            # ORT result.
            tensor_x_new, tensor_y_new, tensor_z_new = run(optimized_copy_copy_copy, [-1.0, 2.0])

            torch.testing.assert_close(tensor_x, tensor_x_new)
            torch.testing.assert_close(tensor_y, tensor_y_new)
            torch.testing.assert_close(tensor_z, tensor_z_new)

        run_to_copy()

    def test_aten_full(self):
        torch._dynamo.reset()

        def run_no_input_model():
            # A function to test.
            def no_input_model():
                return torch.ops.aten.full([2, 3], 1.5)

            @torch._dynamo.optimize("onnxrt")
            def optimized_no_input_model():
                return no_input_model()

            def run(fun):
                tensor_x = fun()
                return tensor_x

            # Baseline.
            tensor_x = run(no_input_model)
            # ORT result.
            tensor_x_new = run(optimized_no_input_model)

            torch.testing.assert_close(tensor_x, tensor_x_new)

        for _ in range(5):
            run_no_input_model()

    def test_aten_full_with_device(self):
        torch._dynamo.reset()

        def run_no_input_model():
            # A function to test.
            def no_input_model():
                return torch.ops.aten.full([2, 3], 1.5, device="cpu")

            optimized_no_input_model = torch.compile(no_input_model, backend="onnxrt")

            def run(fun):
                tensor_x = fun()
                return tensor_x

            # Baseline.
            tensor_x = run(no_input_model)
            # ORT result.
            tensor_x_new = run(optimized_no_input_model)

            torch.testing.assert_close(tensor_x, tensor_x_new)

        for _ in range(5):
            run_no_input_model()

    def test_mnist_model(self):
        torch._dynamo.reset()
        """Test DORT with a simple nn.Module."""

        def run_mnist_model():
            class MNISTModel(nn.Module):
                def __init__(self):
                    super().__init__()
                    self.conv1 = nn.Conv2d(1, 32, 3, 1, bias=False)
                    self.conv2 = nn.Conv2d(32, 64, 3, 1, bias=False)
                    self.fc1 = nn.Linear(9216, 128, bias=False)
                    self.fc2 = nn.Linear(128, 10, bias=False)

                def forward(self, tensor_x: torch.Tensor):
                    tensor_x = self.conv1(tensor_x)
                    tensor_x = F.sigmoid(tensor_x)
                    tensor_x = self.conv2(tensor_x)
                    tensor_x = F.sigmoid(tensor_x)
                    tensor_x = F.max_pool2d(tensor_x, 2)
                    tensor_x = torch.flatten(tensor_x, 1)
                    tensor_x = self.fc1(tensor_x)
                    tensor_x = F.sigmoid(tensor_x)
                    tensor_x = self.fc2(tensor_x)
                    output = F.log_softmax(tensor_x, dim=1)
                    return output

            def run(model, tensor_x, tensor_y):
                tensor_x = tensor_x.detach()
                tensor_y = tensor_y.detach()
                for param in model.parameters():
                    param.grad = None
                output = model(tensor_x)
                loss = F.nll_loss(output, tensor_y)
                # return loss
                loss.backward()
                return loss, (param.grad for param in model.parameters())

            # Input.
            tensor_x = torch.rand((64, 1, 28, 28), dtype=torch.float32)
            # Label.
            tensor_y = torch.randint(0, 9, (64,), dtype=torch.int64)
            model = MNISTModel()

            # Baseline.
            loss, grads = run(model, tensor_x, tensor_y)
            # ORT result.
            local_backend = make_local_backend(dynamic=False, use_aot_autograd=True)
            compiled_model = torch.compile(model, backend=local_backend, dynamic=False)
            loss_new, grads_new = run(compiled_model, tensor_x, tensor_y)

            print(f"MNIST loss: {loss} (pytorch), {loss_new} (ort).")
            torch.testing.assert_close(loss, loss_new, rtol=1e-2, atol=1e-5)
            for grad, grad_new in zip(grads, grads_new, strict=False):
                torch.testing.assert_close(grad, grad_new)

        # Run 5 times because ORT runs have side effects and we want to make sure
        # the code is correct with them.
        for _ in range(5):
            run_mnist_model()


if __name__ == "__main__":
    unittest.main()



import onnx
from onnx import TensorProto, helper


# Since NNAPI EP does not support dynamic shape input and we now switch from the approach of immediately rejecting
# the whole graph in NNAPI EP if it has a dynamic input to checking the dynamic shape at individual operator support check level,
# We have a separated test here using a graph with dynamic input that becomes fixed after a Resize
# Please see BaseOpBuilder::HasSupportedInputs in <repo_root>/onnxruntime/core/providers/nnapi/nnapi_builtin/builders/op_support_checker.cc
def GenerateModel(model_name):  # noqa: N802
    nodes = [
        helper.make_node(
            "Resize",
            ["X", "", "", "Resize_1_sizes"],
            ["Resize_1_output"],
            "resize_1",
            mode="cubic",
        ),
        helper.make_node("Add", ["Resize_1_output", "Add_2_input"], ["Y"], "add"),
    ]

    initializers = [
        helper.make_tensor("Resize_1_sizes", TensorProto.INT64, [4], [1, 1, 3, 3]),
        helper.make_tensor(
            "Add_2_input",
            TensorProto.FLOAT,
            [1, 1, 3, 3],
            [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
        ),
    ]

    inputs = [
        helper.make_tensor_value_info("X", TensorProto.FLOAT, ["1", "1", "N", "N"]),  # used dim_param here
    ]

    outputs = [
        helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 1, 3, 3]),
    ]

    graph = helper.make_graph(nodes, "EP_Dynamic_Graph_Input_Test", inputs, outputs, initializers)

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if __name__ == "__main__":
    GenerateModel("ep_dynamic_graph_input_test.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

from __future__ import annotations

import torch


def torch_nvtx_range_push(msg):
    if hasattr(torch.cuda.nvtx, "range_push"):
        torch.cuda.nvtx.range_push(msg)


def torch_nvtx_range_pop():
    if hasattr(torch.cuda.nvtx, "range_pop"):
        torch.cuda.nvtx.range_pop()


def nvtx_function_decorator(func):
    """Function decorator to record the start and end of NVTX range."""

    def wrapped_fn(*args, **kwargs):
        torch_nvtx_range_push(func.__qualname__)
        ret_val = func(*args, **kwargs)
        torch_nvtx_range_pop()
        return ret_val

    return wrapped_fn


def log_memory_usage(cur_phase: str, rank_0_only=True, step_info="", logger=None, module=None):
    """Log memory usage for the current phase.
    Args:
        cur_phase (str): The current phase.
        rank_0_only (bool, optional): Only log the memory usage for rank 0. Defaults to True.
        step_info (str, optional): The step information. Defaults to "".
        logger (logging.Logger, optional): The logger to log the memory usage. Defaults to None, which means print to stdout.
        module (torch.nn.Module, optional): The module to get parameter, buffer and grad sizes. Defaults to None.
    """
    rank = 0
    if rank_0_only is True:
        if torch.distributed.is_initialized():
            rank = torch.distributed.get_rank()
        if rank != 0:
            return

    _normalizer_factor = float(1024 * 1024)
    _normalizer_unit = "MiB"

    def _normalize(mem_size_in_bytes: float | int) -> str:
        return f"{float(mem_size_in_bytes) / _normalizer_factor:.0f}"

    cur_mem_allocated = _normalize(torch.cuda.memory_allocated())
    max_mem_allocated = _normalize(torch.cuda.max_memory_allocated())
    cur_mem_cached = _normalize(torch.cuda.memory_reserved())
    max_mem_cached = _normalize(torch.cuda.max_memory_reserved())
    torch_mem_stat = torch.cuda.memory_stats()
    cur_mem_inactive = _normalize(torch_mem_stat.get("inactive_split_bytes.all.current", 0))
    max_mem_inactive = _normalize(torch_mem_stat.get("inactive_split_bytes.all.peak", 0))

    mem_stats = [
        ["phase", cur_phase],
        ["allocated", cur_mem_allocated],  # current memory allocated for tensors
        ["max allocated", max_mem_allocated],  # peak memory allocated for tensors
        ["cached", cur_mem_cached],  # current memory cached for the caching allocator
        ["max cached", max_mem_cached],  # peak memory cached for caching allocator.
        ["inactive", cur_mem_inactive],  # amount of inactive, non-releasable memory
        ["max inactive", max_mem_inactive],  # peak of inactive, non-releasable memory
    ]

    # Calculate the total size of parameters and gradients in the model
    if module:
        param_total_size = 0
        grad_total_size = 0
        for p in module.parameters():
            if p.is_cuda:
                param_total_size += p.numel() * p.element_size()
            if p.grad is not None and p.grad.is_cuda:
                grad_total_size += p.grad.numel() * p.grad.element_size()

        # Calculate the total size of buffers in the model
        buffer_total_size = 0
        for b in module.buffers():
            if b.is_cuda:
                buffer_total_size += b.numel() * b.element_size()

        mem_stats.extend(
            [
                ["param", _normalize(param_total_size)],
                ["grad", _normalize(grad_total_size)],
                ["buffer", _normalize(buffer_total_size)],
            ]
        )

    summ = f"rank-{rank} {step_info} memory ({_normalizer_unit})"
    for stat in mem_stats:
        summ += f" | {stat[0]}: {stat[1]}"

    if logger is None:
        print(summ)
    else:
        logger.info(summ)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""

.. _l-example-profiling:

Profile the execution of a simple model
=======================================

*ONNX Runtime* can profile the execution of the model.
This example shows how to interpret the results.
"""

import numpy
import onnx

import onnxruntime as rt
from onnxruntime.datasets import get_example


def change_ir_version(filename, ir_version=6):
    "onnxruntime==1.2.0 does not support opset <= 7 and ir_version > 6"
    with open(filename, "rb") as f:
        model = onnx.load(f)
    model.ir_version = 6
    if model.opset_import[0].version <= 7:
        model.opset_import[0].version = 11
    return model


#########################
# Let's load a very simple model and compute some prediction.

example1 = get_example("mul_1.onnx")
onnx_model = change_ir_version(example1)
onnx_model_str = onnx_model.SerializeToString()
sess = rt.InferenceSession(onnx_model_str, providers=rt.get_available_providers())
input_name = sess.get_inputs()[0].name

x = numpy.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=numpy.float32)
res = sess.run(None, {input_name: x})
print(res)

#########################
# We need to enable to profiling
# before running the predictions.

options = rt.SessionOptions()
options.enable_profiling = True
sess_profile = rt.InferenceSession(onnx_model_str, options, providers=rt.get_available_providers())
input_name = sess.get_inputs()[0].name

x = numpy.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=numpy.float32)

sess.run(None, {input_name: x})
prof_file = sess_profile.end_profiling()
print(prof_file)

###########################
# The results are stored un a file in JSON format.
# Let's see what it contains.
import json  # noqa: E402

with open(prof_file) as f:
    sess_time = json.load(f)
import pprint  # noqa: E402

pprint.pprint(sess_time)



import onnx
from onnx import TensorProto, helper

# graph with a QDQ MatMul node unit where one input is and initializer -> DQ and the other is on a path that
# contains a supported node followed by an unsupported node followed by the DQ -> MatMul.
# The DQ of the initializer is prior to the unsupported node. If the partitioning utils do not process the QDQ node
# unit together, the DQ for the initializer and the first supported node will be in the first partition, which
# incorrectly breaks up the QDQ node unit.
graph_proto = helper.make_graph(
    [
        # DQ of initializer for MatMul B input
        helper.make_node(
            "DequantizeLinear",
            inputs=["matmul_b_uint8", "scale0"],
            outputs=["dq_matmul_b"],
            name="dq_matmul_b",
        ),
        # Treat as supported
        helper.make_node(
            "Mul",
            inputs=["input:0", "scale_input"],
            outputs=["mul:0"],
            name="mul0",
        ),
        # Treat as unsupported
        helper.make_node("Cast", inputs=["mul:0"], outputs=["mul_uint8"], name="cast0", to=2),
        # DQ of MatMul A input
        helper.make_node(
            "DequantizeLinear",
            inputs=["mul_uint8", "scale1"],
            outputs=["dq_matmul_a"],
            name="dq_matmul_a",
        ),
        # MatMul
        helper.make_node(
            "MatMul",
            inputs=[
                "dq_matmul_a",
                "dq_matmul_b",
            ],
            outputs=["matmul_ab"],
            name="matmul_ab",
        ),
        # Q
        helper.make_node(
            "QuantizeLinear",
            inputs=["matmul_ab", "scale2"],
            outputs=["q_matmul_ab"],
            name="q_matmul_ab",
        ),
        # DQ for model output
        helper.make_node(
            "DequantizeLinear",
            inputs=["q_matmul_ab", "scale2"],
            outputs=["out:0"],
            name="dq_graph_output",
        ),
    ],
    "Main_graph",
    [
        helper.make_tensor_value_info("input:0", TensorProto.FLOAT, [3, 2]),
    ],
    [
        helper.make_tensor_value_info("out:0", TensorProto.FLOAT, [3, 2]),
    ],
    [
        helper.make_tensor("scale0", TensorProto.FLOAT, [1], [20.0]),
        helper.make_tensor("scale1", TensorProto.FLOAT, [1], [30.0]),
        helper.make_tensor("scale2", TensorProto.FLOAT, [1], [40.0]),
        helper.make_tensor("matmul_b_uint8", TensorProto.UINT8, [2, 2], [1, 2, 3, 4]),
        helper.make_tensor("scale_input", TensorProto.FLOAT, [2], [3.0, 4.0]),
    ],
)

model = helper.make_model(graph_proto)
onnx.checker.check_model(model, True)
onnx.save(model, "ort_github_issue_19590.onnx")



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import statistics
from dataclasses import dataclass
from enum import Enum
from time import perf_counter

import numpy
import torch
from onnx import TensorProto, helper

from onnxruntime import InferenceSession
from onnxruntime.transformers.io_binding_helper import CudaSession

torch.manual_seed(0)


class GroupNormOpType(Enum):
    GROUP_NORM = 1
    SKIP_GROUP_NORM = 2


@dataclass
class GroupNormConfig:
    batch_size: int
    height: int
    width: int
    channels: int
    epsilon: float = 1e-5
    num_groups: int = 32
    activation: bool = False
    channels_last: bool = True
    fp16: bool = False

    op_type: GroupNormOpType = GroupNormOpType.GROUP_NORM
    has_bias: bool = False
    has_add_out: bool = False
    broadcast_skip: int = 0  # 2 for (N, C), 4 for (N, 1, 1, C)

    def get_skip_symbolic_shape(self):
        skip_shape = {0: ["N", "H", "W", "C"], 2: ["N", "C"], 4: ["N", 1, 1, "C"]}
        return skip_shape[self.broadcast_skip]

    def get_skip_shape(self):
        skip_shape = {
            0: [self.batch_size, self.height, self.width, self.channels],
            2: [self.batch_size, self.channels],
            4: [self.batch_size, 1, 1, self.channels],
        }
        return skip_shape[self.broadcast_skip]

    def broadcast(self, skip: torch.Tensor):
        if self.broadcast_skip == 2:
            return skip.reshape(self.batch_size, 1, 1, self.channels)

        return skip

    @staticmethod
    def create(
        b: int,
        h: int,
        w: int,
        c: int,
        fp16: bool = False,
        activation: bool = False,
        template: int = 0,
        num_groups: int = 32,
    ):
        if template == 0:
            return GroupNormConfig(
                b, h, w, c, fp16=fp16, activation=activation, op_type=GroupNormOpType.GROUP_NORM, num_groups=num_groups
            )

        if template == 1:
            return GroupNormConfig(
                b,
                h,
                w,
                c,
                fp16=fp16,
                activation=activation,
                op_type=GroupNormOpType.SKIP_GROUP_NORM,
                has_bias=True,
                has_add_out=True,
                broadcast_skip=0,
                num_groups=num_groups,
            )

        if template == 2:
            return GroupNormConfig(
                b,
                h,
                w,
                c,
                fp16=fp16,
                activation=activation,
                op_type=GroupNormOpType.SKIP_GROUP_NORM,
                has_bias=False,
                has_add_out=False,
                broadcast_skip=2,
                num_groups=num_groups,
            )

        if template == 3:
            return GroupNormConfig(
                b,
                h,
                w,
                c,
                fp16=fp16,
                activation=activation,
                op_type=GroupNormOpType.SKIP_GROUP_NORM,
                has_bias=True,
                has_add_out=False,
                broadcast_skip=4,
                num_groups=num_groups,
            )

        if template == 4:  # No bias
            return GroupNormConfig(
                b,
                h,
                w,
                c,
                fp16=fp16,
                activation=activation,
                op_type=GroupNormOpType.SKIP_GROUP_NORM,
                has_bias=False,
                has_add_out=True,
                broadcast_skip=0,
                num_groups=num_groups,
            )

        if template == 5:  # No bias, no add_out
            return GroupNormConfig(
                b,
                h,
                w,
                c,
                fp16=fp16,
                activation=activation,
                op_type=GroupNormOpType.SKIP_GROUP_NORM,
                has_bias=False,
                has_add_out=False,
                broadcast_skip=0,
                num_groups=num_groups,
            )

        return None


def create_group_norm_graph(config: GroupNormConfig) -> bytes:
    inputs = ["input", "gamma", "beta"]
    outputs = ["output"]
    op_type = "GroupNorm"
    if config.op_type == GroupNormOpType.SKIP_GROUP_NORM:
        op_type = "SkipGroupNorm"
        inputs = [*inputs, "skip"]
        if config.has_bias:
            inputs = [*inputs, "bias"]
        if config.has_add_out:
            outputs = [*outputs, "add_out"]

    nodes = [
        helper.make_node(
            op_type,
            inputs,
            outputs,
            op_type + "_0",
            activation=int(config.activation),
            channels_last=int(config.channels_last),
            epsilon=config.epsilon,
            groups=config.num_groups,
            domain="com.microsoft",
        ),
    ]

    float_type = TensorProto.FLOAT16 if config.fp16 else TensorProto.FLOAT

    input_shapes = [
        helper.make_tensor_value_info("input", float_type, ["N", "H", "W", "C"]),
        helper.make_tensor_value_info("gamma", TensorProto.FLOAT, ["C"]),
        helper.make_tensor_value_info("beta", TensorProto.FLOAT, ["C"]),
    ]
    output_shapes = [
        helper.make_tensor_value_info("output", float_type, ["N", "H", "W", "C"]),
    ]

    if config.op_type == GroupNormOpType.SKIP_GROUP_NORM:
        input_shapes = [
            *input_shapes,
            helper.make_tensor_value_info("skip", float_type, config.get_skip_symbolic_shape()),
        ]
        if config.has_bias:
            input_shapes = [*input_shapes, helper.make_tensor_value_info("bias", float_type, ["C"])]
        if config.has_add_out:
            output_shapes = [*output_shapes, helper.make_tensor_value_info("add_out", float_type, ["N", "H", "W", "C"])]

    graph = helper.make_graph(
        nodes,
        "Group_Norm_Graph",
        input_shapes,
        output_shapes,
    )

    model = helper.make_model(graph)
    return model.SerializeToString()


def group_norm_ort(
    src: torch.Tensor,
    gamma: torch.Tensor,
    beta: torch.Tensor,
    skip: torch.Tensor | None,
    bias: torch.Tensor | None,
    config: GroupNormConfig,
    measure_latency=False,
) -> tuple[torch.Tensor, torch.Tensor | None, float | None]:
    onnx_model_str = create_group_norm_graph(config)
    ort_session = InferenceSession(onnx_model_str, providers=["CUDAExecutionProvider"])

    session = CudaSession(ort_session, device=torch.device("cuda:0"))

    io_shape = {
        "input": [config.batch_size, config.height, config.width, config.channels],
        "gamma": [config.channels],
        "beta": [config.channels],
        "output": [config.batch_size, config.height, config.width, config.channels],
    }

    if config.op_type == GroupNormOpType.SKIP_GROUP_NORM:
        io_shape["skip"] = config.get_skip_shape()
        if config.has_bias:
            io_shape["bias"] = [config.channels]
        if config.has_add_out:
            io_shape["add_out"] = [config.batch_size, config.height, config.width, config.channels]

    session.allocate_buffers(io_shape)

    ort_inputs = {
        "input": src,
        "gamma": gamma,
        "beta": beta,
    }

    if config.op_type == GroupNormOpType.SKIP_GROUP_NORM:
        ort_inputs["skip"] = skip
        if config.has_bias:
            ort_inputs["bias"] = bias

    ort_outputs = session.infer(ort_inputs)
    output = ort_outputs["output"]

    add_out = (
        ort_outputs["add_out"] if config.op_type == GroupNormOpType.SKIP_GROUP_NORM and config.has_add_out else None
    )

    if measure_latency:
        latency_list = []
        for _ in range(10000):
            start_time = perf_counter()
            session.infer(ort_inputs)
            end_time = perf_counter()
            latency_list.append(end_time - start_time)
        average_latency = statistics.mean(latency_list)
        return output, add_out, average_latency

    return output, add_out, None


def group_norm_torch(
    src: torch.Tensor,
    gamma: torch.Tensor,
    beta: torch.Tensor,
    skip: torch.Tensor | None,
    bias: torch.Tensor | None,
    config: GroupNormConfig,
) -> tuple[torch.Tensor, torch.Tensor | None]:
    add_out = src

    if skip is not None:
        assert config.op_type == GroupNormOpType.SKIP_GROUP_NORM
        add_out = add_out + config.broadcast(skip)

    if bias is not None:
        assert config.op_type == GroupNormOpType.SKIP_GROUP_NORM
        add_out = add_out + bias.reshape(1, 1, 1, bias.shape[0])

    x = add_out
    if config.channels_last:
        x = add_out.clone().permute(0, 3, 1, 2)  # from NHWC to NCHW

    weight = gamma.to(x.dtype)
    bias = beta.to(x.dtype)
    output = torch.nn.functional.group_norm(x, config.num_groups, weight=weight, bias=bias, eps=config.epsilon)

    if config.activation:
        torch.nn.functional.silu(output, inplace=True)

    if config.channels_last:
        output = output.permute(0, 2, 3, 1)  # from NCHW to NHWC

    return output, add_out


def print_tensor(name, tensor):
    # Print in the format that could be directly added to unit tests in C++.
    torch.set_printoptions(precision=6, sci_mode=False, linewidth=100, profile="full", threshold=1000)
    print(name)
    if tensor is not None:
        print("shape", tensor.shape)
        text = str(tensor.clone().flatten())
        print(text.replace("[", "[\n").replace("]", ",\n]").replace(",", "f,"))
    else:
        print(tensor)


def run_parity(config, measure_latency=True, verbose=False):
    float_type = torch.float16 if config.fp16 else torch.float32

    input_tensor = torch.randn(
        config.batch_size,
        config.height,
        config.width,
        config.channels,
        device="cuda",
        dtype=float_type,
        requires_grad=False,
    )

    gamma = torch.randn(
        config.channels,
        device="cuda",
        dtype=torch.float32,
        requires_grad=False,
    )

    beta = torch.randn(
        config.channels,
        device="cuda",
        dtype=torch.float32,
        requires_grad=False,
    )

    skip = None
    bias = None
    if config.op_type == GroupNormOpType.SKIP_GROUP_NORM:
        skip = torch.randn(
            *config.get_skip_shape(),
            device="cuda",
            dtype=float_type,
            requires_grad=False,
        )
        if config.has_bias:
            bias = torch.randn(
                config.channels,
                device="cuda",
                dtype=float_type,
                requires_grad=False,
            )

    if verbose:
        print(config)
        print_tensor("input", input_tensor)
        print_tensor("gamma", gamma)
        print_tensor("beta", beta)
        print_tensor("skip", skip)
        print_tensor("bias", bias)

    out_ort, ort_add_out, latency = group_norm_ort(
        input_tensor, gamma, beta, skip, bias, config, measure_latency=measure_latency
    )

    if verbose:
        print_tensor("out_ort", out_ort)
        print_tensor("ort_add_out", ort_add_out)

    torch_out, torch_add_out = group_norm_torch(input_tensor, gamma, beta, skip, bias, config)

    if verbose:
        print_tensor("torch_out", torch_out)
        print_tensor("torch_add_out", torch_add_out)

    average_diff = numpy.mean(numpy.abs(out_ort.detach().cpu().numpy() - torch_out.detach().cpu().numpy()))

    is_close = numpy.allclose(
        out_ort.detach().cpu().numpy(),
        torch_out.detach().cpu().numpy(),
        rtol=1e-1 if config.fp16 else 1e-3,
        atol=1e-1 if config.fp16 else 1e-3,
        equal_nan=True,
    )

    is_add_out_close = (
        numpy.allclose(
            ort_add_out.detach().cpu().numpy(),
            torch_add_out.detach().cpu().numpy(),
            rtol=1e-1 if config.fp16 else 1e-3,
            atol=1e-1 if config.fp16 else 1e-3,
            equal_nan=True,
        )
        if ort_add_out is not None
        else ""
    )

    # Compare results
    print(
        config.op_type.name,
        " B:",
        config.batch_size,
        " H:",
        config.height,
        " W:",
        config.width,
        " C:",
        config.channels,
        " G:",
        config.num_groups,
        " activation:",
        int(config.activation),
        " channels_last:",
        int(config.channels_last),
        " fp16:",
        int(config.fp16),
        f" Latency(μs): {int(latency * 1e6)}" if isinstance(latency, float) else "",
        " AvgDiff:",
        average_diff,
        " Pass:",
        is_close,
        is_add_out_close,
    )


def get_latent_height_width():
    default_size = [(512, 512), (768, 768), (1024, 1024)]
    small_img_size = [(512, 768), (768, 512)]
    xl_img_size = [
        (1152, 896),
        (896, 1152),
        (1216, 832),
        (832, 1216),
        (1344, 768),
        (768, 1344),
        (1536, 640),
        (640, 1536),
    ]
    return [(int(h / 8), int(w / 8)) for (h, w) in default_size + small_img_size + xl_img_size]


def get_channels():
    return [128, 256, 512, 1024, 2048, 320, 640, 960, 1920, 2560, 384, 768, 1536, 3072, 1152, 2304]


def run_activation(template: int, fp16, measure_latency=False):
    print("Test GroupNorm with Silu Activation for ", "fp16" if fp16 else "fp32")
    for b in [2]:
        for h, w in get_latent_height_width():
            for c in get_channels():
                config = GroupNormConfig.create(b, h, w, c, fp16=fp16, activation=True, template=template)
                run_parity(config, measure_latency=measure_latency)


def run_no_activation(template: int, fp16, measure_latency=False):
    print("Test GroupNorm without Activation for ", "fp16" if fp16 else "fp32")
    for b in [1, 2, 4]:
        for h, w in get_latent_height_width():
            for c in get_channels():
                config = GroupNormConfig.create(b, h, w, c, fp16=fp16, template=template)
                run_parity(config, measure_latency=measure_latency)


def run_all_groups(template: int, fp16, measure_latency=False):
    group_sizes = [1, 2, 4, 8, 16, 32]
    print("Test GroupNorm for different group sizes:", group_sizes)
    for group_size in group_sizes:
        for h, w in get_latent_height_width()[:3]:
            for c in get_channels()[:2]:
                config = GroupNormConfig.create(2, h, w, c, fp16=fp16, num_groups=group_size, template=template)
                run_parity(config, measure_latency=measure_latency)


def run_odd_channels(template: int, fp16, measure_latency=False):
    # Test some random number of channels that can be divisible by 2 * num_groups
    for h, w in get_latent_height_width():
        for c in [448, 704, 832, 1664, 2240, 2688, 2880, 3008]:
            config = GroupNormConfig.create(2, h, w, c, fp16=fp16, num_groups=32, template=template)
            run_parity(config, measure_latency=measure_latency)


def run_small_inputs(template: int, fp16):
    config = GroupNormConfig.create(2, 2, 2, 16, fp16=fp16, activation=False, num_groups=4, template=template)
    run_parity(config, measure_latency=False)

    config = GroupNormConfig.create(1, 1, 1, 64, fp16=fp16, activation=False, num_groups=8, template=template)
    run_parity(config, measure_latency=False)

    config = GroupNormConfig.create(1, 1, 1, 64, fp16=fp16, activation=True, num_groups=8, template=template)
    run_parity(config, measure_latency=False)


def run_performance(fp16):
    # Run perf test to tune parameters for given number of channels.
    for h, w in get_latent_height_width()[:3]:
        for c in get_channels():
            config = GroupNormConfig.create(2, h, w, c, fp16=fp16, num_groups=32, template=0)
            run_parity(config, measure_latency=True)


def run_all(template: int):
    for fp16 in [True, False]:
        run_small_inputs(template, fp16)
        run_odd_channels(template, fp16)
        run_all_groups(template, fp16)
        run_activation(template, fp16)
        run_no_activation(template, fp16)


def run_not_implemented():
    # Expect failure. Check whether the error message is expected.
    try:
        config = GroupNormConfig(1, 2, 2, 513, num_groups=3)
        run_parity(config)
    except RuntimeError as e:
        assert "GroupNorm in CUDA does not support the input: n=1 h=2 w=2 c=513 groups=3" in str(e)


def main():
    run_performance(True)

    run_not_implemented()

    for template in range(6):
        run_all(template)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import unittest

import numpy as np
import onnx
from onnx import TensorProto, helper
from op_test_utils import TestDataFeeds, check_model_correctness, check_op_type_count

from onnxruntime.quantization import quantize_dynamic


class TestOpEmbedLayerNormalization(unittest.TestCase):
    def input_feeds_int32(self, n, name2shape):
        input_data_list = []
        for _i in range(n):
            inputs = {}
            for name, shape in name2shape.items():
                inputs.update({name: np.ones(shape).astype(np.int32)})
            input_data_list.extend([inputs])

        dr = TestDataFeeds(input_data_list)
        return dr

    def construct_model(self, batch, hidden_size, sequence_length, model_path, empty_segment=False):
        #    <segment_ids>    <input_ids>
        #              \        /
        #       (EmbedLayerNormalization)
        #               /       \
        # <layernorm_output>  <mask_index_output>

        # Inputs to EmbedLayerNormalizationNode
        input_ids_shape = [batch, sequence_length]
        input_ids_tensor = helper.make_tensor_value_info("input_ids", TensorProto.INT32, input_ids_shape)

        segment_ids_shape = [batch, sequence_length]
        segment_ids_tensor = helper.make_tensor_value_info("segment_ids", TensorProto.INT32, segment_ids_shape)

        # EmbedLayerNormalization Node Constants and Weights:
        word_embed_shape = [32, hidden_size]
        word_embed_weights = np.random.random_sample(word_embed_shape).astype(dtype="float32")
        word_embed_initializer = onnx.numpy_helper.from_array(word_embed_weights, name="word_embed")

        pos_embed_shape = [16, hidden_size]
        pos_embed_weights = np.random.random_sample(pos_embed_shape).astype(dtype="float32")
        pos_embed_initializer = onnx.numpy_helper.from_array(pos_embed_weights, name="pos_embed")

        seg_embed_shape = [2, hidden_size]
        seg_embed_weights = np.random.random_sample(seg_embed_shape).astype(dtype="float32")
        seg_embed_initializer = onnx.numpy_helper.from_array(seg_embed_weights, name="seg_embed")

        gamma_shape = [hidden_size]
        gamma = np.random.random_sample(gamma_shape).astype(dtype="float32")
        gamma_initializer = onnx.numpy_helper.from_array(gamma, name="gamma")

        beta_shape = [hidden_size]
        beta = np.random.random_sample(beta_shape).astype(dtype="float32")
        beta_initializer = onnx.numpy_helper.from_array(beta, name="beta")

        # EmbedLayerNormalization Outputs:
        layernorm_out_shape = [batch, sequence_length, hidden_size]
        layernorm_out_tensor = helper.make_tensor_value_info("layernorm_out", TensorProto.FLOAT, layernorm_out_shape)

        mask_index_out_shape = [batch]
        mask_index_out_tensor = helper.make_tensor_value_info("mask_index_out", TensorProto.INT32, mask_index_out_shape)

        # EmbedLayerNormalization Node:
        embed_layer_norm_inputs = [
            "input_ids",
            "segment_ids" if not empty_segment else "",
            "word_embed",
            "pos_embed",
            "seg_embed" if not empty_segment else "",
            "gamma",
            "beta",
        ]
        embed_layer_norm_outputs = ["layernorm_out", "mask_index_out"]
        embed_layer_norm_node = helper.make_node(
            "EmbedLayerNormalization",
            embed_layer_norm_inputs,
            embed_layer_norm_outputs,
            domain="com.microsoft",
        )

        # Construct the Graph and Model:
        nodes = [embed_layer_norm_node]
        graph_name = "embed_layernorm_graph"
        inputs = [input_ids_tensor, segment_ids_tensor]
        outputs = [layernorm_out_tensor, mask_index_out_tensor]
        initializers = (
            [
                word_embed_initializer,
                pos_embed_initializer,
                seg_embed_initializer,
                gamma_initializer,
                beta_initializer,
            ]
            if not empty_segment
            else [word_embed_initializer, pos_embed_initializer, gamma_initializer, beta_initializer]
        )

        graph = helper.make_graph(nodes, graph_name, inputs, outputs, initializer=initializers)
        model = helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", 14), helper.make_opsetid("com.microsoft", 1)]
        )
        model.ir_version = 7  # use stable onnx ir version
        onnx.save(model, model_path)

    def test_quantize_batch_size_1(self):
        batch = 1
        hidden_size = 4
        sequence_length = 4

        model_f32_path = "test_embed_layer_norm_unit_test_batch1.onnx"
        model_uint8_path = "test_embed_layer_norm_unit_test_batch1_uint8.onnx"

        self.construct_model(batch, hidden_size, sequence_length, model_f32_path)

        data_reader = self.input_feeds_int32(
            1,
            {
                "input_ids": [batch, sequence_length],
                "segment_ids": [batch, sequence_length],
            },
        )

        quantize_dynamic(model_f32_path, model_uint8_path)

        # Quantization should not have any DequantizeLinear nodes:
        qnode_counts = {"DequantizeLinear": 0, "QEmbedLayerNormalization": 1}
        check_op_type_count(self, model_uint8_path, **qnode_counts)
        data_reader.rewind()

        check_model_correctness(self, model_f32_path, model_uint8_path, data_reader.get_next())

    def test_quantize_batch_size_1_empty_segment(self):
        batch = 1
        hidden_size = 4
        sequence_length = 4

        model_f32_path = "test_embed_layer_norm_unit_test_batch1_empty_segment.onnx"
        model_uint8_path = "test_embed_layer_norm_unit_test_batch1_uint8_empty_segment.onnx"

        self.construct_model(batch, hidden_size, sequence_length, model_f32_path, empty_segment=True)

        data_reader = self.input_feeds_int32(
            1,
            {
                "input_ids": [batch, sequence_length],
                "segment_ids": [batch, sequence_length],
            },
        )

        quantize_dynamic(model_f32_path, model_uint8_path)

        # Quantization should not have any DequantizeLinear nodes:
        qnode_counts = {"DequantizeLinear": 0, "QEmbedLayerNormalization": 1}
        check_op_type_count(self, model_uint8_path, **qnode_counts)
        data_reader.rewind()

        check_model_correctness(self, model_f32_path, model_uint8_path, data_reader.get_next())

    def test_quantize_batch_size_2(self):
        batch = 2
        hidden_size = 4
        sequence_length = 4

        model_f32_path = "test_embed_layer_norm_unit_test_batch2.onnx"
        model_uint8_path = "test_embed_layer_norm_unit_test_batch2_uint8.onnx"

        self.construct_model(batch, hidden_size, sequence_length, model_f32_path)

        data_reader = self.input_feeds_int32(
            1,
            {
                "input_ids": [batch, sequence_length],
                "segment_ids": [batch, sequence_length],
            },
        )

        quantize_dynamic(model_f32_path, model_uint8_path)

        # Quantization should not have any DequantizeLinear nodes:
        qnode_counts = {"DequantizeLinear": 0, "QEmbedLayerNormalization": 1}
        check_op_type_count(self, model_uint8_path, **qnode_counts)
        data_reader.rewind()

        check_model_correctness(self, model_f32_path, model_uint8_path, data_reader.get_next())

    def test_quantize_batch_size_2_empty_segment(self):
        batch = 2
        hidden_size = 4
        sequence_length = 4

        model_f32_path = "test_embed_layer_norm_unit_test_batch2_empty_segment.onnx"
        model_uint8_path = "test_embed_layer_norm_unit_test_batch2_uint8_empty_segment.onnx"

        self.construct_model(batch, hidden_size, sequence_length, model_f32_path, empty_segment=True)

        data_reader = self.input_feeds_int32(
            1,
            {
                "input_ids": [batch, sequence_length],
                "segment_ids": [batch, sequence_length],
            },
        )

        quantize_dynamic(model_f32_path, model_uint8_path)

        # Quantization should not have any DequantizeLinear nodes:
        qnode_counts = {"DequantizeLinear": 0, "QEmbedLayerNormalization": 1}
        check_op_type_count(self, model_uint8_path, **qnode_counts)
        data_reader.rewind()

        check_model_correctness(self, model_f32_path, model_uint8_path, data_reader.get_next())


if __name__ == "__main__":
    unittest.main()



import onnx


def find_node(graph_proto, op_type):
    nodes = []
    map_input_node = {}
    for node in graph_proto.node:
        if node.op_type == op_type:
            map_input_node[node.input[0]] = node
            if op_type == "Div" or op_type == "Mul":
                map_input_node[node.input[1]] = node
            nodes.append(node)
    return nodes, map_input_node


def gen_attribute(key, value):
    attr = AttributeProto()  # noqa: F821
    attr.name = key
    attr.ints.extend(int(v) for v in value)
    attr.type = AttributeProto.INTS  # noqa: F821
    return attr


def layer_norm_transform(model_proto):
    # a layer norm subgraph
    # input
    #   |
    # ReduceMean
    #  __|____
    # |       |
    # Sub     Sub
    # |       |
    # |       Pow
    # |        |
    # |        ReduceMean
    # |        |
    # |        Add
    # |        |
    # |__    __Sqrt
    #    |  |
    #     Div
    #     |
    #     Mul
    #     |
    #     Add
    #     |
    #     output

    graph_proto = model_proto.graph

    _, map_input_Div = find_node(graph_proto, "Div")  # noqa: N806

    _, map_input_Sqrt = find_node(graph_proto, "Sqrt")  # noqa: N806

    _, map_input_Add = find_node(graph_proto, "Add")  # noqa: N806

    nodes_ReduceMean, map_input_ReduceMean = find_node(graph_proto, "ReduceMean")  # noqa: N806

    _, map_input_Pow = find_node(graph_proto, "Pow")  # noqa: N806

    _, map_input_Mul = find_node(graph_proto, "Mul")  # noqa: N806

    # find right side Sub (see the layer norm subgrapg)
    nodes_Sub = []  # noqa: N806
    map_input_Sub = {}  # noqa: N806
    for node in graph_proto.node:
        if node.op_type == "Sub":
            if node.output[0] in map_input_Pow:
                nodes_Sub.append(node)
                map_input_Sub[node.input[1]] = node

    # find first ReduceMean
    first_ReduceMean = []  # noqa: N806
    first_ReduceMean_outputs = []  # noqa: N806
    for node in nodes_ReduceMean:
        if node.output[0] in map_input_Sub:
            first_ReduceMean.append(node)
            first_ReduceMean_outputs.append(node.output[0])

    # find constant node
    nodes_Constant = []  # noqa: N806
    map_output_Constant = {}  # noqa: N806
    for node in graph_proto.node:
        if node.op_type == "Constant":
            nodes_Constant.append(node)
            map_output_Constant[node.output[0]] = node

    id = 0
    removed_nodes = []
    layer_norm_nodes = []
    # Replace with layer norm
    for node in first_ReduceMean:
        layer_norm_input = []
        layer_norm_output = []
        layer_norm_input.append(node.input[0])

        # collect nodes within a layer norm subgraph.
        # skip building layer norm node if there is a pattern miss-match.
        if node.output[0] not in map_input_Sub:
            continue

        node_sub = map_input_Sub[node.output[0]]
        if node_sub.output[0] not in map_input_Pow:
            continue

        node_pow = map_input_Pow[node_sub.output[0]]
        if node_pow.output[0] not in map_input_ReduceMean:
            continue

        node_reduce = map_input_ReduceMean[node_pow.output[0]]
        if node_reduce.output[0] not in map_input_Add:
            continue

        node_Add = map_input_Add[node_reduce.output[0]]  # noqa: N806
        if node_Add.output[0] not in map_input_Sqrt:
            continue

        node_Sqrt = map_input_Sqrt[node_Add.output[0]]  # noqa: N806
        if node_Sqrt.output[0] not in map_input_Div:
            continue

        node_Div = map_input_Div[node_Sqrt.output[0]]  # noqa: N806
        if node_Div.output[0] not in map_input_Mul:
            continue

        node_Mul = map_input_Mul[node_Div.output[0]]  # noqa: N806

        if node_Mul.input[0] != node_Div.output[0]:
            layer_norm_input.append(node_Mul.input[0])
        else:
            layer_norm_input.append(node_Mul.input[1])

        if node_Mul.output[0] not in map_input_Add:
            continue

        node_Add1 = map_input_Add[node_Mul.output[0]]  # noqa: N806
        layer_norm_input.append(node_Add1.input[1])

        removed_nodes.append(node)
        removed_nodes.append(node_sub)
        removed_nodes.append(node_pow)
        removed_nodes.append(node_reduce)
        removed_nodes.append(node_Add)
        removed_nodes.append(node_Sqrt)
        removed_nodes.append(node_Div)
        removed_nodes.append(node_Mul)
        removed_nodes.append(node_Add1)
        removed_nodes.append(map_output_Constant[node_pow.input[1]])

        removed_nodes.append(map_output_Constant[node_Add.input[1]])
        layer_norm_output.append(node_Add1.output[0])
        id = id + 1
        layer_norm_output.append("saved_mean_" + str(id))
        id = id + 1
        layer_norm_output.append("saved_inv_std_var_" + str(id))
        layer_norm = onnx.helper.make_node(
            "LayerNormalization",
            layer_norm_input,
            layer_norm_output,
            "LayerNormalization_" + str(id),
            None,
            axis=node_reduce.attribute[0].ints[0],
            epsilon=9.999999960041972e-13,
        )
        layer_norm_nodes.append(layer_norm)

    # remove left side Subs
    for node in graph_proto.node:
        if node.op_type == "Sub":
            if node.input[1] in first_ReduceMean_outputs:
                removed_nodes.append(node)

    all_nodes = []
    for node in graph_proto.node:
        if node not in removed_nodes:
            all_nodes.append(node)

    for node in layer_norm_nodes:
        all_nodes.append(node)  # noqa: PERF402

    graph_proto.ClearField("node")
    graph_proto.node.extend(all_nodes)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import copy
import os
import time

import torch
from transformers import BartConfig, BartForConditionalGeneration, file_utils
from utils import export_helper

from onnxruntime import InferenceSession, SessionOptions


def decoder_config_update(config: BartConfig):
    """
    Add parameters into decoder config to help exporting. These parameters
    are later consumed with control flow and assertion

    Args:
        config: BART config
    """

    new_attributes = {
        "is_decoder_exported": False,
        "is_decoder_with_past_exported": False,
        "during_export": False,
        "expected_args": {"use_cache": True, "return_dict": True},
        "expected_inputs": ["decoder_input_ids", "attention_mask", "encoder_outputs"],
    }
    config.update(new_attributes)
    return config


class DecoderWrapper(torch.nn.Module):
    """main part of BART Decoder wrapper.

    We split BART into three parts: Encoder plus initial part of decodoer, main part of decodoer, and Beam Search.
    This module is the main part of decodoer.

    Attributes:
        model: the BART model.
    """

    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(
        self,
        decoder_input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        *past: torch.Tensor,
    ):
        encoder_outputs = file_utils.ModelOutput()
        encoder_outputs["last_hidden_state"] = encoder_hidden_states
        encoder_outputs["hidden_states"] = None
        encoder_outputs["attentions"] = None
        if len(past) == 0:
            past_key_values = None
        else:
            past_key_values = export_helper.back_group_by_layer(past)
        decoder_out = self.model(
            None,
            encoder_outputs=encoder_outputs,
            decoder_input_ids=decoder_input_ids,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            use_cache=True,
            return_dict=True,
        )
        present_self, _ = export_helper.group_by_self_and_cross(decoder_out.past_key_values)
        return decoder_out.logits, present_self


def _decoder_forward_wrapper(model: BartForConditionalGeneration, decoder_config: BartConfig, args):
    """Wrapper function to pass in args and config.

    Exporting the decoder of BART by inserting torch.onnx.export and wrapper
    torch.nn.Module into `forward`.

    Args:
        model: BartForConditionalGeneration module.
        args: User input.
        decoder_config: BART config after `decoder_config_update`.
    """
    model._forward = model.forward

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, **kwargs):
        # attention_mask is required to show as forward parameters based on `generation_utils.py`
        # reference: https://github.com/huggingface/transformers/blob/main/src/transformers/generation_utils.py#L1200-L1203
        kwargs["attention_mask"] = attention_mask

        if decoder_config.is_decoder_exported and decoder_config.is_decoder_with_past_exported:
            outputs = self._forward(input_ids, **kwargs)
            return outputs

        inputs = []
        for k, v in decoder_config.expected_args.items():
            assert k in kwargs, f"Expecting {k} in kwargs."
            assert v == kwargs[k], f"For {k}, expected {v}, got {kwargs[k]}."

        for k in decoder_config.expected_inputs:
            assert k in kwargs
            if k == "encoder_outputs":
                inputs.append(kwargs[k]["last_hidden_state"])
            elif kwargs[k] is not None:
                inputs.append(kwargs[k].type(torch.int32))

        # delete unused key//value
        for k in list(kwargs.keys()):
            if (
                k not in decoder_config.expected_args
                and k not in decoder_config.expected_inputs
                and k != "past_key_values"
            ):
                del kwargs[k]

        encoder_outputs = kwargs["encoder_outputs"]
        for k, v in encoder_outputs.items():
            if k == "last_hidden_state":
                assert v is not None
            else:
                assert v is None

        assert "past_key_values" in kwargs
        # need to preserve past inputs, because values are altered after forward.
        past_inputs = kwargs["past_key_values"]
        past_input_list = []
        if past_inputs is not None and not decoder_config.is_decoder_with_past_exported:
            past_input_list = copy.deepcopy(export_helper.group_by_self_and_cross(past_inputs, concat=True))

        # compare `decoder_out_pt` with ORT results.
        decoder_out_pt = self._forward(input_ids, **kwargs)
        past_outputs = decoder_out_pt.past_key_values

        if past_inputs is not None and decoder_config.is_decoder_with_past_exported:
            return decoder_out_pt
        if past_inputs is None and decoder_config.is_decoder_exported:
            return decoder_out_pt

        input_past_names = []
        onnx_model_path = os.path.join(args.output, "decoder.onnx")

        if past_inputs is not None:
            decoder_config.is_decoder_with_past_exported = True
            inputs.extend(past_input_list)
            input_past_names = export_helper.get_input_names(past_inputs, encoder=False)
            onnx_model_path = os.path.join(args.output, "decoder_past.onnx")
        else:
            decoder_config.is_decoder_exported = True

        input_names = ["input_ids", "encoder_attention_mask", "encoder_hidden_states", *input_past_names]
        output_past_names = export_helper.get_output_names(past_outputs)
        output_names = ["logits", *output_past_names]

        sequence_length = "1"
        num_heads = str(decoder_config.encoder_attention_heads)
        head_size = str(decoder_config.encoder_attention_heads)
        hidden_size = str(decoder_config.d_model)

        dynamic_axes = {
            "input_ids": {0: "batch", 1: sequence_length},
            "encoder_attention_mask": {0: "batch", 1: "sequence"},
            "encoder_hidden_states": {0: "batch", 1: "sequence", 2: hidden_size},
            "logits": {0: "batch", 1: sequence_length},
        }

        for name in output_past_names:
            if "self" in name:
                dynamic_axes[name] = {
                    0: "batch",
                    1: num_heads,
                    2: sequence_length,
                    3: head_size,
                }

        for name in input_past_names:
            if "self" in name:
                dynamic_axes[name] = {
                    0: "batch",
                    1: num_heads,
                    2: "sequence",
                    3: head_size,
                }

        wrapped_decoder = DecoderWrapper(model).eval()
        # export decoder with past
        if decoder_config.is_decoder_with_past_exported:
            logits, present = wrapped_decoder(inputs[0], inputs[1], inputs[2], *inputs[3:])
            decoder_config.during_export = True
            torch.onnx.export(
                wrapped_decoder,
                tuple(inputs),
                onnx_model_path,
                opset_version=args.opset_version,
                do_constant_folding=False,
                input_names=input_names,
                output_names=output_names,
                dynamic_axes=dynamic_axes,
                verbose=True,
            )
            decoder_config.during_export = False

            # Test the generated model with onnxruntime
            print("========== ORT inference test on Decoder ... ==========")
            ort_inputs = {name: value.cpu().numpy() for name, value in zip(input_names, inputs, strict=False)}
            # NOTE: encoder_hidden_states is not used and deleted
            ort_inputs.pop("encoder_hidden_states")
            sess_options = SessionOptions()
            sess_options.log_severity_level = 4
            sess = InferenceSession(onnx_model_path, sess_options, providers=["CPUExecutionProvider"])
            out = sess.run(None, ort_inputs)

            for ort_out, torch_out in zip(out, [logits, *present], strict=False):
                torch.testing.assert_close(ort_out, torch_out.cpu().numpy(), check_dtype=True, atol=1e-4, rtol=1e-2)

            print("========== [SUCCESS] ORT inference test on Decoder ==========")

        return decoder_out_pt

    return forward


def export_decoder(args):
    """Export BART decoder to ONNX.

    By replacing the inner function of `BartForConditionalGeneration` forward,
    we export encoder model into ONNX with the usage of model.genration()

    Args:
        args: User input.

    Return:
        Decoder ONNX model in the given output directory, or under onnx_models folder.
    """
    beam = args.num_beams
    min_length = args.min_length
    max_length = args.max_length
    repetition_penalty = args.repetition_penalty
    no_repeat_ngram_size = args.no_repeat_ngram_size

    config, tokenizer = export_helper.initialize_config(args)
    config.use_decoder = True
    config = decoder_config_update(config)

    with torch.no_grad():
        model, input_data = export_helper.initialize_model(config, tokenizer, args)
        start_time = time.time()

        model.forward = _decoder_forward_wrapper(model, config, args).__get__(model, BartForConditionalGeneration)

        pred_ids = model.generate(
            input_data,
            decoder_start_token_id=tokenizer.eos_token_id,
            num_beams=beam,
            num_return_sequences=beam,
            min_length=min_length,
            max_length=max_length,
            repetition_penalty=repetition_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            use_cache=True,
        )
        time_cost = time.time() - start_time
        print(f"--- {time_cost} seconds ---")
        print(tokenizer.decode(pred_ids[0], skip_special_tokens=True, clean_up_tokenization_spaces=False))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import unittest

import coloredlogs
import pytest
from parity_utilities import find_transformers_source

from onnxruntime import get_available_providers

if find_transformers_source(sub_dir_paths=["models", "gpt2"]):
    from models.gpt2.convert_to_onnx import main as gpt2_to_onnx
else:
    from onnxruntime.transformers.models.gpt2.convert_to_onnx import main as gpt2_to_onnx


class TestGpt2ConvertToOnnx(unittest.TestCase):
    """Test convert_to_onnx.py of GPT-2 model."""

    def setUp(self):
        self.test_cuda = "CUDAExecutionProvider" in get_available_providers()

    def run_gpt2_to_onnx(self, arguments: str, stage: int):
        result = gpt2_to_onnx(arguments.split())
        self.assertIsNotNone(result)
        self.assertTrue(isinstance(result, dict))
        self.assertIn("top1_match_rate", result)
        self.assertGreater(result["top1_match_rate"], 0.98)
        self.assertIn("stage", result)
        self.assertEqual(result["stage"], stage)
        if stage == 1:
            self.assertIn("average_latency(batch_size=8,sequence_length=32,past_sequence_length=0)", result)
        else:
            self.assertIn("average_latency(batch_size=8,sequence_length=1,past_sequence_length=32)", result)

    @pytest.mark.slow
    def test_stage1(self):
        """Test stage 1 onnx model"""
        if self.test_cuda:
            self.run_gpt2_to_onnx("-m distilgpt2 -p fp32 --use_gpu -s 1 -t 100 --overwrite", 1)

    @pytest.mark.slow
    def test_stage2(self):
        """Test stage 2 onnx model"""
        if self.test_cuda:
            self.run_gpt2_to_onnx("-m distilgpt2 -p fp32 --use_gpu -s 2 -t 100 --overwrite", 2)

    @pytest.mark.slow
    def test_auto_mixed_precision(self):
        """Test mixed preicison onnx model"""
        if self.test_cuda:
            self.run_gpt2_to_onnx("-m distilgpt2 -p fp32 --use_gpu -p fp16 -o -a -t 100 --overwrite", 0)


if __name__ == "__main__":
    coloredlogs.install(fmt="%(message)s")
    logging.getLogger("transformers").setLevel(logging.ERROR)
    unittest.main()



import numpy as np
import onnx
from onnx import OperatorSetIdProto, TensorProto, helper, numpy_helper

vocab_size = 256  # 30258

X = helper.make_tensor_value_info("input", TensorProto.FLOAT, ["batch", "seqlen", 128])
unsqueezed_masked_lm_positions = helper.make_tensor_value_info(
    "unsqueezed_masked_lm_positions",
    TensorProto.INT64,
    ["batch", "dynamic_prediction_count", 1],
)
Y = helper.make_tensor_value_info("output", TensorProto.FLOAT, ["batch", "dynamic_prediction_count", vocab_size])
Gather_Y = helper.make_tensor_value_info("gather_output", TensorProto.FLOAT, ["batch", 128])

layer_norm1_weight_np_vals = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
layer_norm1_weight_initializer = numpy_helper.from_array(
    layer_norm1_weight_np_vals, "bert.encoder.layer.2.output.LayerNorm.weight"
)

layer_norm1_bias_np_vals = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
layer_norm1_bias_initializer = numpy_helper.from_array(
    layer_norm1_bias_np_vals, "bert.encoder.layer.2.output.LayerNorm.bias"
)

matmul1_np_vals = np.random.uniform(0.0, 1.0, (128, 128)).astype(np.float32).reshape((128, 128))
matmul1_initializer = numpy_helper.from_array(matmul1_np_vals, "matmul1_initializer")

add1_np_vals = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
add1_initializer = numpy_helper.from_array(add1_np_vals, "add1_initializerr")

layer_norm2_weight_np_vals = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
layer_norm2_weight_initializer = numpy_helper.from_array(
    layer_norm2_weight_np_vals, "cls.predictions.transform.LayerNorm.weight"
)

layer_norm2_bias_np_vals = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
layer_norm2_bias_initializer = numpy_helper.from_array(
    layer_norm2_bias_np_vals, "cls.predictions.transform.LayerNorm.bias"
)

matmul2_np_vals = np.random.uniform(0.0, 1.0, (128, vocab_size)).astype(np.float32).reshape((128, vocab_size))
matmul2_initializer = numpy_helper.from_array(matmul2_np_vals, "bert.embeddings.word_embeddings.weight_transposed")

add2_np_vals = np.random.uniform(0.0, 1.0, (vocab_size)).astype(np.float32).reshape(vocab_size)
add2_initializer = numpy_helper.from_array(add2_np_vals, "cls.predictions.bias")

gather_indice_np_vals = np.asarray([0]).astype(np.int64).reshape(())
gather_indice_initializer = numpy_helper.from_array(gather_indice_np_vals, "gather_indice_initializer")

nodes = []
layer_norm1 = helper.make_node(
    "LayerNormalization",
    ["input", layer_norm1_weight_initializer.name, layer_norm1_bias_initializer.name],
    ["layer_norm1", "saved_mean1", "saved_inv_std_var1"],
    name="layer_norm_1",
    epsilon=9.999999960041972e-13,
    axis=-1,
)
nodes.append(layer_norm1)

gather1 = helper.make_node(
    "Gather",
    ["layer_norm1", gather_indice_initializer.name],
    ["gather_output"],
    name="gather_output",
    axis=1,
)
nodes.append(gather1)

matmul1 = helper.make_node("MatMul", ["layer_norm1", matmul1_initializer.name], ["matmul1"], name="matmul_1")
nodes.append(matmul1)

add1 = helper.make_node("Add", [add1_initializer.name, "matmul1"], ["add1"], name="add_1")
nodes.append(add1)

gelu1 = helper.make_node("Gelu", ["add1"], ["gelu1"], name="gelu_1", domain="com.microsoft")
nodes.append(gelu1)

layer_norm2 = helper.make_node(
    "LayerNormalization",
    ["gelu1", layer_norm2_weight_initializer.name, layer_norm2_bias_initializer.name],
    ["layer_norm2", "saved_mean2", "saved_inv_std_var2"],
    name="layer_norm_2",
    epsilon=9.999999960041972e-13,
    axis=-1,
)
nodes.append(layer_norm2)

matmul2 = helper.make_node("MatMul", ["layer_norm2", matmul2_initializer.name], ["matmul2"], name="matmul_2")
nodes.append(matmul2)

add2 = helper.make_node("Add", ["matmul2", add2_initializer.name], ["add2"], name="add_2")
nodes.append(add2)

gathernd1 = helper.make_node(
    "GatherND",
    ["add2", "unsqueezed_masked_lm_positions"],
    ["gathernd1"],
    name="gathernd_1",
    batch_dims=1,
)
nodes.append(gathernd1)

identity1 = helper.make_node("Identity", ["gathernd1"], ["output"], name="output")
nodes.append(identity1)

initializers = [
    layer_norm1_weight_initializer,
    layer_norm1_bias_initializer,
    matmul1_initializer,
    add1_initializer,
    layer_norm2_weight_initializer,
    layer_norm2_bias_initializer,
    matmul2_initializer,
    add2_initializer,
    gather_indice_initializer,
]
# Create the graph (GraphProto)
graph_def = helper.make_graph(
    nodes,
    "test-model",
    [X, unsqueezed_masked_lm_positions],
    [Y, Gather_Y],
    initializers,
)

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"

opsets.append(msdomain)
kwargs = {}
kwargs["opset_imports"] = opsets

model_def = helper.make_model(graph_def, producer_name="onnx-example", **kwargs)

onnx.save(model_def, "computation_reduction_transformer.onnx")



# This test script is a modified version of Pytorch's tutorial.
# For details, see https://pytorch.org/tutorials/intermediate/ddp_tutorial.html.
import argparse
import os
import sys  # noqa: F401
import tempfile

import torch
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.nn as nn
import torch.optim as optim
from torch.nn.parallel import DistributedDataParallel as DDP  # noqa: N817

import onnxruntime  # noqa: F401
from onnxruntime.training.ortmodule import ORTModule


def setup(rank, world_size):
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = "12355"

    # initialize the process group
    dist.init_process_group("gloo", rank=rank, world_size=world_size)


def cleanup():
    dist.destroy_process_group()


class ToyModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.net1 = nn.Linear(10, 10)
        self.relu = nn.ReLU()
        self.net2 = nn.Linear(10, 5)

    def forward(self, x):
        return self.net2(self.relu(self.net1(x)))


def demo_basic(rank, world_size, use_ort_module):
    torch.manual_seed(0)
    print(f"Running basic DDP example on rank {rank}.")
    setup(rank, world_size)

    # create model and move it to GPU with id rank
    model = ToyModel().to(rank)
    if use_ort_module:
        model = ORTModule(model)
        print(f"  Rank {rank} uses ORTModule.")
    else:
        print(f"  Rank {rank} uses Pytorch's nn.Module.")

    ddp_model = DDP(model, device_ids=[rank])

    loss_fn = nn.MSELoss()
    optimizer = optim.Adagrad(ddp_model.parameters(), lr=0.01)

    x = torch.randn(20, 10).to(rank)
    y = torch.randn(20, 5).to(rank)

    loss_history = []

    for i in range(5):
        optimizer.zero_grad()
        p = ddp_model(x)
        loss = loss_fn(p, y)
        with torch.no_grad():
            print(f"  Rank {rank} at iteration {i} has loss {loss}.")
        loss.backward()
        optimizer.step()
        with torch.no_grad():
            loss_history.append(torch.unsqueeze(loss, 0))

    loss_history = torch.cat(loss_history).cpu()
    expected_loss_history = torch.FloatTensor(
        [1.4909229278564453, 1.432194471359253, 1.39592707157135, 1.367714762687683, 1.3445055484771729]
    )

    assert torch.allclose(expected_loss_history, loss_history)

    cleanup()


def demo_checkpoint(rank, world_size, use_ort_module):
    torch.manual_seed(rank)
    print(f"Running DDP checkpoint example on rank {rank}.")
    setup(rank, world_size)

    if use_ort_module:
        print(f"  Rank {rank} uses ORTModule.")
        model = ToyModel().to(rank)
        model = ORTModule(model)
    else:
        print(f"  Rank {rank} uses Pytorch's nn.Module.")
        model = ToyModel().to(rank)

    ddp_model = DDP(model, device_ids=[rank])

    loss_fn = nn.MSELoss()
    optimizer = optim.SGD(ddp_model.parameters(), lr=0.001)

    CHECKPOINT_PATH = os.path.join(tempfile.gettempdir(), "model.checkpoint")  # noqa: N806
    if rank == 0:
        # All processes should see same parameters as they all start from same
        # random parameters and gradients are synchronized in backward passes.
        # Therefore, saving it in one process is sufficient.
        torch.save(ddp_model.state_dict(), CHECKPOINT_PATH)

    # Use a barrier() to make sure that process 1 loads the model after process
    # 0 saves it.
    dist.barrier()
    # configure map_location properly
    map_location = {"cuda:0": f"cuda:{rank}"}
    ddp_model.load_state_dict(torch.load(CHECKPOINT_PATH, map_location=map_location))

    optimizer.zero_grad()
    outputs = ddp_model(torch.randn(20, 10))
    labels = torch.randn(20, 5).to(rank)
    loss_fn = nn.MSELoss()
    loss = loss_fn(outputs, labels)
    loss.backward()
    optimizer.step()

    print(f"Rank {rank} sees loss {loss}")

    if rank == 0:
        assert torch.allclose(loss.cpu(), torch.FloatTensor([1.4909229278564453]))
    elif rank == 1:
        assert torch.allclose(loss.cpu(), torch.FloatTensor([1.0177688598632812]))
    elif rank == 2:
        assert torch.allclose(loss.cpu(), torch.FloatTensor([1.290669322013855]))
    elif rank == 3:
        assert torch.allclose(loss.cpu(), torch.FloatTensor([0.825118362903595]))
    else:
        raise AssertionError()

    # Not necessary to use a dist.barrier() to guard the file deletion below
    # as the AllReduce ops in the backward pass of DDP already served as
    # a synchronization.

    if rank == 0:
        os.remove(CHECKPOINT_PATH)

    cleanup()


"""
CustomEmbedding is adapted from
https://github.com/huggingface/transformers/blob/312b104ff65514736c0475814fec19e47425b0b5/src/transformers/models/distilbert/modeling_distilbert.py#L91.
"""


class CustomEmbeddings(nn.Module):
    def __init__(self):
        super().__init__()
        vocab_size = 511
        dim = 10
        pad_token_id = 0
        max_position_embeddings = 16
        self.word_embeddings = nn.Embedding(vocab_size, dim, padding_idx=pad_token_id)
        self.position_embeddings = nn.Embedding(max_position_embeddings, dim)
        self.LayerNorm = nn.LayerNorm(dim, eps=1e-12)
        self.register_buffer("position_ids", torch.arange(max_position_embeddings).expand((1, -1)), persistent=False)

    def forward(self, input_ids: torch.Tensor) -> torch.Tensor:
        input_embeds = self.word_embeddings(input_ids)  # (bs, max_seq_length, dim)
        seq_length = input_embeds.size(1)
        position_ids = self.position_ids[:, :seq_length]
        position_embeddings = self.position_embeddings(position_ids)  # (bs, max_seq_length, dim)

        embeddings = input_embeds + position_embeddings  # (bs, max_seq_length, dim)
        embeddings = self.LayerNorm(embeddings)  # (bs, max_seq_length, dim)
        return embeddings


"""
This module calls `CustomEmbeddings`, which will generate a series of nodes (Shape->Gather->Unsqueeze...) whos
allocate output tensors on CPU memory. Then this test can converge failures when CUDA EP enabled, CPU device
allocation and usage are correct.
"""


class AnotherLayerOfToyModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.embedding = CustomEmbeddings()
        self.t = ToyModel()

    def forward(self, x):
        embed_val = self.embedding(x)
        return self.t(embed_val)


"""
`Mixed device allocation` here means ORT backend allocates output tensors on CPU for some nodes and
on CUDA for other nodes. This test could help catch regression when ORT allocation planner logic got changed with bugs.
"""


def demo_mixed_device_allocation_training(rank, world_size, use_ort_module):
    torch.manual_seed(0)
    print(f"Running basic DDP example on rank {rank}.")
    setup(rank, world_size)
    device = "cuda:" + str(rank)

    # create a model and move it to GPU with id rank
    model = AnotherLayerOfToyModel().to(device)
    if use_ort_module:
        model = ORTModule(model)
        print(f"  Rank {rank} uses ORTModule.")
    else:
        print(f"  Rank {rank} uses Pytorch's nn.Module.")

    ddp_model = DDP(model, device_ids=[device])

    loss_fn = nn.MSELoss()
    optimizer = optim.Adagrad(ddp_model.parameters(), lr=0.01)

    batch = 2
    max_seq_length = 16
    x = torch.randint(1, 511, (batch, max_seq_length)).to(device)
    y = torch.randn(batch, max_seq_length, 5).to(device)

    loss_history = []

    for i in range(5):
        optimizer.zero_grad()
        p = ddp_model(x)
        loss = loss_fn(p, y)
        with torch.no_grad():
            print(f"  Rank {rank} at iteration {i} has loss {loss}.")
        loss.backward()
        optimizer.step()
        with torch.no_grad():
            loss_history.append(torch.unsqueeze(loss, 0))

    loss_history = torch.cat(loss_history).cpu()
    expected_loss_history = torch.FloatTensor([1.1589857340, 1.0975260735, 1.0628030300, 1.0386666059, 1.0196533203])

    assert torch.allclose(expected_loss_history, loss_history)

    cleanup()


def run_demo(demo_fn, world_size, use_ort_module):
    mp.spawn(demo_fn, args=(world_size, use_ort_module), nprocs=world_size, join=True)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--use_ort_module", action="store_true")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    run_demo(demo_basic, 4, args.use_ort_module)
    run_demo(demo_checkpoint, 4, args.use_ort_module)
    run_demo(demo_mixed_device_allocation_training, 4, args.use_ort_module)



import onnx
from onnx import TensorProto, helper


def GenerateModel(model_name):  # noqa: N802
    initializers = [
        helper.make_tensor(
            "weight",
            TensorProto.FLOAT,
            [1, 4],
            [1.0, 2.0, 3.0, 4.0],
        ),
    ]

    nodes = [
        helper.make_node(
            "custom op",
            ["input1", "weight"],
            ["output"],
            "custom op",
            domain="test.customop",
        ),
    ]

    graph = helper.make_graph(
        nodes,
        "custom op graph",
        [  # input
            helper.make_tensor_value_info("input1", TensorProto.FLOAT, [1, 4]),
        ],
        [  # output
            helper.make_tensor_value_info("output", TensorProto.FLOAT, [1, 4]),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if __name__ == "__main__":
    GenerateModel("test_kernel_info_get_const_input.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
"""
Convert a Bert large model exported by Keras2Onnx to a tiny model for test purpose.
The input model is generated like the following (need install keras2onnx from source):

    import numpy
    import keras2onnx
    from transformers import (TFBertForQuestionAnswering, BertTokenizer)

    tokenizer = BertTokenizer.from_pretrained'bert-large-uncased-whole-word-masking-finetuned-squad', do_lower_case=True, cache_dir=cache_dir)
    model = TFBertForQuestionAnswering.from_pretrained(model_name_or_path, cache_dir=cache_dir)

    question, text = "What is ONNX Runtime?", "ONNX Runtime is a performance-focused inference engine for ONNX models."
    inputs = tokenizer.encode_plus(question, text, add_special_tokens=True, return_tensors='tf')

    output_model_path =  os.path.join(output_dir, 'keras_{}.onnx'.format(model_name_or_path))
    if not os.path.exists(output_model_path):
        model.predict(inputs)
        onnx_model = keras2onnx.convert_keras(model, model.name)
        keras2onnx.save_model(onnx_model, output_model_path)
"""

import argparse
import os
import random
import sys  # noqa: F401
import timeit
from pathlib import Path

import numpy as np
import onnx
import onnx.utils
from onnx import ModelProto, TensorProto, numpy_helper
from onnxruntime_tools.transformers.onnx_model import OnnxModel

import onnxruntime

DICT_SIZE = 20
SEQ_LEN = 7
""" This class creates a tiny bert model for test purpose. """


class TinyBertOnnxModel(OnnxModel):
    def __init__(self, model, verbose):
        super().__init__(model, verbose)
        self.resize_model()

    def resize_weight(self, initializer_name, target_shape):
        weight = self.get_initializer(initializer_name)
        w = numpy_helper.to_array(weight)

        target_w = w
        if len(target_shape) == 1:
            target_w = w[: target_shape[0]]
        elif len(target_shape) == 2:
            target_w = w[: target_shape[0], : target_shape[1]]
        elif len(target_shape) == 3:
            target_w = w[: target_shape[0], : target_shape[1], : target_shape[2]]
        else:
            print("at most 3 dimensions")

        tensor = onnx.helper.make_tensor(
            name=initializer_name + "_resize",
            data_type=TensorProto.FLOAT,
            dims=target_shape,
            vals=target_w.flatten().tolist(),
        )

        return tensor

    def resize_model(self):
        graph = self.model.graph
        initializers = graph.initializer

        # parameters of input base model.
        old_parameters = {
            "seq_len": 26,
            "hidden_size": 1024,
            "num_heads": 16,
            "size_per_head": 64,
            "word_dict_size": [28996, 30522],  # list of supported dictionary size.
            "max_word_position": 512,
        }

        # parameters of output tiny model.
        new_parameters = {
            "seq_len": SEQ_LEN,
            "hidden_size": 8,
            "num_heads": 2,
            "size_per_head": 4,
            "word_dict_size": DICT_SIZE,
            "max_word_position": 10,
        }

        for input in graph.input:
            if input.type.tensor_type.shape.dim[1].dim_value == old_parameters["seq_len"]:
                print("input", input.name, input.type.tensor_type.shape)
                input.type.tensor_type.shape.dim[1].dim_value = new_parameters["seq_len"]
                print("=>", input.type.tensor_type.shape)

        reshapes = {}
        for initializer in initializers:
            tensor = numpy_helper.to_array(initializer)
            dtype = np.float32 if initializer.data_type == 1 else np.int32
            if len(tensor.shape) == 1 and tensor.shape[0] == 1:
                if tensor == old_parameters["num_heads"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["num_heads"],
                        "=>[",
                        new_parameters["num_heads"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["num_heads"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["seq_len"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["seq_len"],
                        "=>[",
                        new_parameters["seq_len"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["seq_len"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["size_per_head"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["size_per_head"],
                        "=>[",
                        new_parameters["size_per_head"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["size_per_head"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["hidden_size"],
                        "=>[",
                        new_parameters["hidden_size"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["hidden_size"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 4 * old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        4 * old_parameters["hidden_size"],
                        "=>[",
                        4 * new_parameters["hidden_size"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([4 * new_parameters["hidden_size"]], dtype=dtype),
                            initializer.name,
                        )
                    )
            elif len(tensor.shape) == 0:
                if tensor == old_parameters["num_heads"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["num_heads"],
                        "=>",
                        new_parameters["num_heads"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["num_heads"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["seq_len"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["seq_len"],
                        "=>",
                        new_parameters["seq_len"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["seq_len"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["size_per_head"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["size_per_head"],
                        "=>",
                        new_parameters["size_per_head"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["size_per_head"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["hidden_size"],
                        "=>",
                        new_parameters["hidden_size"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["hidden_size"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 4 * old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        4 * old_parameters["hidden_size"],
                        "=>",
                        4 * new_parameters["hidden_size"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(4 * new_parameters["hidden_size"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 1.0 / np.sqrt(old_parameters["size_per_head"]):
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        1.0 / np.sqrt(old_parameters["size_per_head"]),
                        "=>",
                        1.0 / np.sqrt(new_parameters["size_per_head"]),
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(
                                1.0 / np.sqrt(new_parameters["size_per_head"]),
                                dtype=dtype,
                            ),
                            initializer.name,
                        )
                    )

            new_shape = []
            shape_changed = False
            for dim in tensor.shape:
                if dim == old_parameters["hidden_size"]:
                    new_shape.append(new_parameters["hidden_size"])
                    shape_changed = True
                elif dim == 4 * old_parameters["hidden_size"]:
                    new_shape.append(4 * new_parameters["hidden_size"])
                    shape_changed = True
                elif dim in old_parameters["word_dict_size"]:
                    new_shape.append(new_parameters["word_dict_size"])
                    shape_changed = True
                elif dim == old_parameters["max_word_position"]:
                    new_shape.append(new_parameters["max_word_position"])
                    shape_changed = True
                else:
                    new_shape.append(dim)
            if shape_changed:
                reshapes[initializer.name] = new_shape
                print("initializer", initializer.name, tensor.shape, "=>", new_shape)

        for initializer_name, reshape_name in reshapes.items():
            self.replace_input_of_all_nodes(initializer_name, initializer_name + "_resize")
            tensor = self.resize_weight(initializer_name, reshape_name)
            self.model.graph.initializer.extend([tensor])

        self.use_dynamic_axes()

    def use_dynamic_axes(self, dynamic_batch_dim="batch_size", seq_len=7):
        """
        Update input and output shape to use dynamic axes.
        """
        for input in self.model.graph.input:
            dim_proto = input.type.tensor_type.shape.dim[0]
            dim_proto.dim_param = dynamic_batch_dim
            dim_proto = input.type.tensor_type.shape.dim[1]
            dim_proto.dim_value = seq_len

        for output in self.model.graph.output:
            dim_proto = output.type.tensor_type.shape.dim[0]
            dim_proto.dim_param = dynamic_batch_dim
            dim_proto = output.type.tensor_type.shape.dim[1]
            dim_proto.dim_value = seq_len


def generate_test_data(
    onnx_file,
    output_path,
    batch_size,
    sequence_length,
    use_cpu=True,
    input_tensor_only=False,
    dictionary_size=DICT_SIZE,
    test_cases=3,
):
    input_data_type = np.int32
    for test_case in range(test_cases):
        input_1 = np.random.randint(dictionary_size, size=(batch_size, sequence_length), dtype=input_data_type)
        tensor_1 = numpy_helper.from_array(input_1, "input_ids")

        actual_seq_len = random.randint(sequence_length - 3, sequence_length)
        input_2 = np.zeros((batch_size, sequence_length), dtype=input_data_type)
        temp = np.ones((batch_size, actual_seq_len), dtype=input_data_type)
        input_2[: temp.shape[0], : temp.shape[1]] = temp
        tensor_2 = numpy_helper.from_array(input_2, "attention_mask")

        input_3 = np.zeros((batch_size, sequence_length), dtype=input_data_type)
        tensor_3 = numpy_helper.from_array(input_3, "token_type_ids")

        path = os.path.join(output_path, "test_data_set_" + str(test_case))
        try:
            os.mkdir(path)
        except OSError:
            print(f"Creation of the directory {path} failed")
        else:
            print(f"Successfully created the directory {path} ")

        if input_tensor_only:
            return

        sess_options = onnxruntime.SessionOptions()
        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
        sess = onnxruntime.InferenceSession(onnx_file, sess_options, providers=["CPUExecutionProvider"])

        output_names = [output.name for output in sess.get_outputs()]
        inputs = {
            "input_ids": input_1,
            "attention_mask": input_2,
            "token_type_ids": input_3,
        }
        print("inputs", inputs)
        result = sess.run(output_names, inputs)

        with open(os.path.join(path, f"input_{0}.pb"), "wb") as f:
            f.write(tensor_1.SerializeToString())
        with open(os.path.join(path, f"input_{1}.pb"), "wb") as f:
            f.write(tensor_2.SerializeToString())
        with open(os.path.join(path, f"input_{2}.pb"), "wb") as f:
            f.write(tensor_3.SerializeToString())

        for i, _output_name in enumerate(output_names):
            tensor_result = numpy_helper.from_array(
                np.asarray(result[i]).reshape((batch_size, sequence_length)),
                output_names[i],
            )
            with open(os.path.join(path, f"output_{i}.pb"), "wb") as f:
                f.write(tensor_result.SerializeToString())

        start_time = timeit.default_timer()

        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED

        path_prefix = onnx_file[:-5]  # remove .onnx suffix
        if use_cpu:
            sess_options.optimized_model_filepath = path_prefix + "_optimized_cpu.onnx"
        else:
            sess_options.optimized_model_filepath = path_prefix + "_optimized_gpu.onnx"

        session = onnxruntime.InferenceSession(
            onnx_file,
            sess_options=sess_options,
            providers=onnxruntime.get_available_providers(),
        )
        if use_cpu:
            session.set_providers(["CPUExecutionProvider"])  # use cpu
        else:
            if "CUDAExecutionProvider" not in session.get_providers():
                print("Warning: GPU not found")
                continue
        outputs = session.run(None, inputs)
        evalTime = timeit.default_timer() - start_time  # noqa: N806
        if outputs[0].tolist() != result[0].tolist():
            print(
                f"Error: not same result after optimization. use_cpu={use_cpu}, no_opt_output={result[0].tolist()}, opt_output={outputs[1].tolist()}"
            )
        print(f"** Evaluation done in total {evalTime} secs")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", required=True, type=str)
    parser.add_argument("--output", required=True, type=str)
    parser.add_argument("--float16", required=False, action="store_true")
    parser.set_defaults(float16=False)
    args = parser.parse_args()

    model = ModelProto()
    with open(args.input, "rb") as f:
        model.ParseFromString(f.read())

    bert_model = TinyBertOnnxModel(model, False)

    if args.float16:
        bert_model.convert_model_float32_to_float16()

    bert_model.update_graph()
    bert_model.remove_unused_constant()

    print("opset version", bert_model.model.opset_import[0].version)

    with open(args.output, "wb") as out:
        out.write(bert_model.model.SerializeToString())

    p = Path(args.output)
    data_path = p.parent

    batch_size = 1
    sequence_length = SEQ_LEN

    generate_test_data(args.output, data_path, batch_size, sequence_length, use_cpu=not args.float16)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""This file is used to generate test data for LR scheduler optimizer tests in
orttraining/orttraining/test/training_api/core/training_api_tests.cc."""

import torch
from torch.optim.lr_scheduler import LambdaLR


class SingleParameterModule(torch.nn.Module):
    """A dummy module containing only one trainable parameter."""

    def __init__(self, input_size, hidden_size):
        super().__init__()
        self.fc1 = torch.nn.Linear(input_size, hidden_size, bias=False)

    def forward(self, input1):
        """Module forward call."""
        out = self.fc1(input1)
        return out


class WarmupLinearSchedule(LambdaLR):
    """Linear warmup and then linear decay.
    Linearly increases learning rate from 0 to 1 over `warmup_steps` training steps.
    Linearly decreases learning rate from 1. to 0. over remaining `t_total - warmup_steps` steps.
    """

    def __init__(self, optimizer, warmup_steps, t_total, last_epoch=-1):
        self.warmup_steps = warmup_steps
        self.t_total = t_total
        super().__init__(optimizer, self.lr_lambda, last_epoch=last_epoch)

    def lr_lambda(self, step):
        print(f"warmup_step_count_: {self.warmup_steps}, step: {step}, total_step_count_: {self.t_total}")
        if step < self.warmup_steps:
            return float(step) / float(max(1, self.warmup_steps))
        return max(0.0, float(self.t_total - step) / float(max(1.0, self.t_total - self.warmup_steps)))


def main():
    """Main entry."""
    num_training_steps = 100
    device = "cuda"
    batch_size, dimension_in, dimension_hidden = 2, 2, 3

    input = torch.randn(batch_size, dimension_in, device=device, dtype=torch.float32)
    save_ckpt_step = num_training_steps // 2
    num_warmup_step_data_dict = {
        "0": 0,
        "30": 30,
        # checkpoint at 50 step.
        "70": 70,
        "200": 200,
    }

    for warmup_name, num_warmup_steps in num_warmup_step_data_dict.items():
        pt_model = SingleParameterModule(dimension_in, dimension_hidden).to(device)

        import tempfile

        fp = tempfile.NamedTemporaryFile()  # noqa: SIM115

        adamw_optimizer = torch.optim.AdamW(pt_model.parameters(), lr=1e-3)
        scheduler = WarmupLinearSchedule(adamw_optimizer, num_warmup_steps, num_training_steps)
        data = []
        for i in range(num_training_steps):
            data.append([scheduler.last_epoch, scheduler.get_last_lr()])
            prediction = pt_model(input)
            loss = prediction.sum()
            loss.backward()
            adamw_optimizer.step()
            adamw_optimizer.zero_grad()
            scheduler.step()

            if i == save_ckpt_step:
                torch.save(
                    {
                        "optimizer": adamw_optimizer.state_dict(),
                        "lr_scheduler": scheduler.state_dict(),
                    },
                    fp.name,
                )

        import json

        json_file_name = f"warmup_linear_scheduler_warmupstep-{warmup_name}.json"
        with open(json_file_name, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

        data = []
        state_dict = torch.load(fp.name)
        new_adamw_optimizer = torch.optim.AdamW(pt_model.parameters(), lr=1e-3)
        new_adamw_optimizer.load_state_dict(state_dict["optimizer"])

        new_scheduler = WarmupLinearSchedule(new_adamw_optimizer, num_warmup_steps, num_training_steps)
        new_scheduler.load_state_dict(state_dict["lr_scheduler"])
        for i in range(save_ckpt_step + 1, num_training_steps):  # noqa: B007
            data.append([new_scheduler.last_epoch, new_scheduler.get_last_lr()])
            prediction = pt_model(input)
            loss = prediction.sum()
            loss.backward()
            new_adamw_optimizer.step()
            new_adamw_optimizer.zero_grad()
            new_scheduler.step()

        import json

        json_file_name = f"warmup_linear_scheduler_warmupstep-{warmup_name}_restored.json"
        with open(json_file_name, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import os

import torch
from packaging.version import Version

_all_optimizers = []

if (
    "ORTMODULE_USE_EFFICIENT_ATTENTION" in os.environ
    and int(os.getenv("ORTMODULE_USE_EFFICIENT_ATTENTION")) == 1
    and Version(torch.__version__) >= Version("2.3.0")
):
    from ._aten_attn import optimize_graph_for_aten_efficient_attention  # noqa: F401

    _all_optimizers.append("optimize_graph_for_aten_efficient_attention")

__all__ = _all_optimizers  # noqa: PLE0605



import os


def get_name(name):
    if os.path.exists(name):
        return name
    rel = os.path.join("testdata", name)
    if os.path.exists(rel):
        return rel
    this = os.path.dirname(__file__)
    data = os.path.join(this, "..", "testdata")
    res = os.path.join(data, name)
    if os.path.exists(res):
        return res
    raise FileNotFoundError(f"Unable to find '{name}' or '{rel}' or '{res}'")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import os
import subprocess
import sys
from glob import glob
from shutil import copyfile

import torch
from packaging import version

from onnxruntime.training import ortmodule


def _list_extensions(path):
    extensions = []
    for root, _, files in os.walk(path):
        for name in files:
            if name.lower() == "setup.py":
                extensions.append(os.path.join(root, name))
    return extensions


def _list_cpu_extensions():
    return _list_extensions(os.path.join(ortmodule.ORTMODULE_TORCH_CPP_DIR, "cpu"))


def _list_cuda_extensions():
    return _list_extensions(os.path.join(ortmodule.ORTMODULE_TORCH_CPP_DIR, "cuda"))


def _install_extension(ext_name, ext_path, cwd):
    ret_code = subprocess.call((sys.executable, ext_path, "build"), cwd=cwd)
    if ret_code != 0:
        print(f"There was an error compiling '{ext_name}' PyTorch CPP extension")
        sys.exit(ret_code)


def _get_cuda_extra_build_params():
    nvcc_extra_args = ["-lineinfo", "-O3", "--use_fast_math"]
    cuda_version = torch.version.cuda
    if cuda_version is not None and version.parse(cuda_version) > version.parse("11.2"):
        # If number is 0, the number of threads used is the number of CPUs on the machine.
        nvcc_extra_args += ["--threads", "0"]

    os.environ["ONNXRUNTIME_CUDA_NVCC_EXTRA_ARGS"] = ",".join(nvcc_extra_args)


def build_torch_cpp_extensions():
    """Builds PyTorch CPP extensions and returns metadata."""
    # Run this from within onnxruntime package folder
    is_gpu_available = (torch.version.cuda is not None or torch.version.hip is not None) and (
        ortmodule.ONNXRUNTIME_CUDA_VERSION is not None or ortmodule.ONNXRUNTIME_ROCM_VERSION is not None
    )

    # Docker build don't have CUDA support, but Torch C++ extensions with CUDA may be forced
    force_cuda = bool(os.environ.get("ONNXRUNTIME_FORCE_CUDA", False))

    os.chdir(ortmodule.ORTMODULE_TORCH_CPP_DIR)

    # Extensions might leverage CUDA/ROCM versions internally
    os.environ["ONNXRUNTIME_CUDA_VERSION"] = (
        ortmodule.ONNXRUNTIME_CUDA_VERSION if ortmodule.ONNXRUNTIME_CUDA_VERSION is not None else ""
    )
    os.environ["ONNXRUNTIME_ROCM_VERSION"] = (
        ortmodule.ONNXRUNTIME_ROCM_VERSION if ortmodule.ONNXRUNTIME_ROCM_VERSION is not None else ""
    )

    if torch.version.cuda is not None and ortmodule.ONNXRUNTIME_CUDA_VERSION is not None:
        _get_cuda_extra_build_params()

    ############################################################################
    # Pytorch CPP Extensions that DO require CUDA/ROCM
    ############################################################################
    if is_gpu_available or force_cuda:
        for ext_setup in _list_cuda_extensions():
            _install_extension(ext_setup.split(os.sep)[-2], ext_setup, ortmodule.ORTMODULE_TORCH_CPP_DIR)

    ############################################################################
    # Pytorch CPP Extensions that DO NOT require CUDA/ROCM
    ############################################################################
    for ext_setup in _list_cpu_extensions():
        _install_extension(ext_setup.split(os.sep)[-2], ext_setup, ortmodule.ORTMODULE_TORCH_CPP_DIR)

    ############################################################################
    # Install Pytorch CPP Extensions into local onnxruntime package folder
    ############################################################################
    torch_cpp_exts = glob(os.path.join(ortmodule.ORTMODULE_TORCH_CPP_DIR, "build", "lib.*", "*.so"))
    torch_cpp_exts.extend(glob(os.path.join(ortmodule.ORTMODULE_TORCH_CPP_DIR, "build", "lib.*", "*.dll")))
    torch_cpp_exts.extend(glob(os.path.join(ortmodule.ORTMODULE_TORCH_CPP_DIR, "build", "lib.*", "*.dylib")))
    for ext in torch_cpp_exts:
        dest_ext = os.path.join(ortmodule.ORTMODULE_TORCH_CPP_DIR, os.path.basename(ext))
        print(f"Installing {ext} -> {dest_ext}")
        copyfile(ext, dest_ext)

    # Tear down
    os.environ.pop("ONNXRUNTIME_CUDA_VERSION")
    os.environ.pop("ONNXRUNTIME_ROCM_VERSION")


if __name__ == "__main__":
    build_torch_cpp_extensions()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import logging
import os  # noqa: F401
import sys

from _test_commons import run_subprocess

logging.basicConfig(format="%(asctime)s %(name)s [%(levelname)s] - %(message)s", level=logging.DEBUG)
log = logging.getLogger("Build")


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--cmd_line_with_args",
        required=True,
        help="command line with arguments to be executed in a subprocess. \
        it expects a single string containing arguments separated by spaces.",
    )
    parser.add_argument("--cwd", help="working directory")
    # parser.add_argument("--env", help="env variables.")
    parser.add_argument("--env", help="env variables", nargs=2, action="append", default=[])

    return parser.parse_args()


launch_args = parse_arguments()

print("sys.executable: ", sys.executable)
cmd_line_with_args = launch_args.cmd_line_with_args.split()
for n, arg in enumerate(cmd_line_with_args):
    if arg == "python":
        cmd_line_with_args[n] = sys.executable

run_subprocess(cmd_line_with_args, cwd=launch_args.cwd, env=dict(launch_args.env), log=log)
